from pydantic import BaseModel, Field
from typing import Optional, List
import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.cms.model.generics import GenericSyncModule, GenericAsyncModule
from altscore.cms.model.common import Money, Schedule, Terms

import datetime as dt


class Balance(BaseModel):
    fees: Optional[Money] = Field(alias="fees")
    interest: Optional[Money] = Field(alias="interest")
    principal: Optional[Money] = Field(alias="principal")
    taxes: Optional[Money] = Field(alias="taxes")
    penalties: Optional[Money] = Field(alias="penalties")
    total: Optional[Money] = Field(alias="total")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Client(BaseModel):
    id: str = Field(alias="clientId")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    partner_id: str = Field(alias="partnerId")
    external_id: str = Field(alias="externalId")
    email: str = Field(alias="email")
    legal_name: str = Field(alias="legalName")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Transaction(BaseModel):
    id: str = Field(alias="transactionId")
    breakdown: Optional[List[Balance]] = Field(alias="breakdown")
    amount: Money = Field(alias="amount")
    type: str = Field(alias="type")
    date: str = Field(alias="date")
    reference_id: str = Field(alias="referenceId", default=None)
    notes: Optional[str] = Field(alias="notes", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class DebtAPIDTO(BaseModel):
    id: str = Field(alias="debtId")
    flow_id: str = Field(alias="flowId")
    tenant: str = Field(alias="tenant")
    reference_id: str = Field(alias="referenceId")
    status: str = Field(alias="status")
    sub_status: str = Field(alias="subStatus")
    client: Client = Field(alias="client")
    balance: Balance = Field(alias="balance")
    closing_balance: Money = Field(alias="closingBalance")
    schedule: List[Schedule] = Field(alias="schedule")
    terms: Terms = Field(alias="terms")
    transactions: List[Transaction] = Field(alias="transactions")
    created_at: str = Field(alias="createdAt")
    updated_at: str = Field(alias="updatedAt")
    version: int = Field(alias="version")
    days_past_due: Optional[int] = Field(alias="daysPastDue", default=None)
    max_days_past_due: Optional[int] = Field(alias="maxDaysPastDue", default=None)
    source: Optional[str] = Field(alias="source", default="simple_credit")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Payment(BaseModel):
    debt_id: str = Field(alias="debtId")
    amount: Money = Field(alias="amount")
    payment_date: str = Field(alias="paymentDate")
    reference_id: str = Field(alias="referenceId")
    notes: Optional[str] = Field(alias="notes")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class PenaltyBalance(BaseModel):
    fees: Money = Field(alias="fees")
    interest: Money = Field(alias="interest")
    principal: Money = Field(alias="principal")
    taxes: Money = Field(alias="taxes")
    penalties: Money = Field(alias="penalties")
    total: Money = Field(alias="total")
    installment: int = Field(alias="installment")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class Penalty(BaseModel):
    amount: Money = Field(alias="amount")
    breakdown: List[PenaltyBalance] = Field(alias="breakdown")
    date: str = Field(alias="date")
    notes: Optional[str] = Field(alias="notes")
    reference_id: str = Field(alias="referenceId")
    transaction_id: str = Field(alias="transactionId")
    type: str = Field(alias="type")


class CreateDebt(BaseModel):
    flow_id: str = Field(alias="flowId")
    disbursed_at: Optional[str] = Field(alias="disbursedAt", default=None)
    amount: Optional[Money] = Field(alias="amount", default=None)

    class Config:
        allow_population_by_field_name = True
        populate_by_alias = True
        populate_by_name = True


class Waiver(BaseModel):
    amount: Money = Field(alias="amount") # This amount should be the sum of the breakdown amounts
    breakdown: Optional[Balance] = Field(alias="breakdown", default=None) # These amount should be negative floats
    notes: Optional[str] = Field(alias="notes", default=None)
    reference_id: str = Field(alias="referenceId")
    waiver_date: str = Field(alias="waiverDate")

    class Config:
        allow_population_by_field_name = True
        populate_by_alias = True
        populate_by_name = True


class DebtBase:

    @staticmethod
    def _payments(debt_id: str):
        return f"/v1/debts/{debt_id}/payments"

    @staticmethod
    def _penalties(debt_id: str):
        return f"/v1/debts/{debt_id}/penalties"

    @staticmethod
    def _waivers(debt_id: str):
        return f"/v1/debts/{debt_id}/waivers"


class DebtAsync(DebtBase):
    data: DebtAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DebtAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    @retry_on_401_async
    async def get_payments(self) -> List[Payment]:
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(
                self._payments(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            return [Payment.parse_obj(e) for e in response.json()]

    @retry_on_401_async
    async def submit_payment(self, amount: str, currency: str, reference_id: str, notes: Optional[str] = None,
                             payment_date: Optional[dt.date] = None) -> None:
        if payment_date is None:
            payment_date = dt.date.today()
        if notes is None:
            notes = ""
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                self._payments(self.data.id),
                json={
                    "amount": {
                        "amount": amount,
                        "currency": currency
                    },
                    "referenceId": reference_id,
                    "notes": notes,
                    "paymentDate": payment_date.strftime("%Y-%m-%d")
                },
                timeout=30,
                headers=self._header_builder()
            )
            raise_for_status_improved(response)

    @retry_on_401_async
    async def get_penalties(self) -> List[Penalty]:
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.get(
                self._waivers(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            return [Penalty.parse_obj(e) for e in response.json()]

    @retry_on_401_async
    async def apply_waiver(self, waiver: dict) -> None:
        waiver = Waiver.parse_obj(waiver)
        async with httpx.AsyncClient(base_url=self.base_url) as client:
            response = await client.post(
                self._waivers(self.data.id),
                headers=self._header_builder(),
                json=waiver.dict(by_alias=True, exclude_none=True),
                timeout=30
            )
            raise_for_status_improved(response)

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class DebtSync(DebtBase):
    data: DebtAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DebtAPIDTO):
        super().__init__()
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    @retry_on_401
    def get_payments(self) -> List[Payment]:
        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(
                self._payments(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            return [Payment.parse_obj(e) for e in response.json()]

    @retry_on_401
    def submit_payment(self, amount: str, currency: str, reference_id: str, notes: Optional[str] = None,
                       payment_date: Optional[dt.date] = None) -> None:
        if payment_date is None:
            payment_date = dt.date.today()
        if notes is None:
            notes = ""
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                self._payments(self.data.id),
                json={
                    "amount": {
                        "amount": amount,
                        "currency": currency
                    },
                    "referenceId": reference_id,
                    "notes": notes,
                    "paymentDate": payment_date.strftime("%Y-%m-%d")
                },
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)

    @retry_on_401
    def get_penalties(self) -> List[Penalty]:
        with httpx.Client(base_url=self.base_url) as client:
            response = client.get(
                self._penalties(self.data.id),
                headers=self._header_builder(),
                timeout=30
            )
            raise_for_status_improved(response)
            return [Penalty.parse_obj(e) for e in response.json()]

    @retry_on_401
    def apply_waiver(self, waiver: dict) -> None:
        waiver = Waiver.parse_obj(waiver)
        with httpx.Client(base_url=self.base_url) as client:
            response = client.post(
                self._waivers(self.data.id),
                headers=self._header_builder(),
                json=waiver.dict(by_alias=True, exclude_none=True),
                timeout=30
            )
            raise_for_status_improved(response)
            return None

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.id})"


class DebtsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            async_resource=DebtAsync,
            retrieve_data_model=DebtAPIDTO,
            create_data_model=None,
            update_data_model=None,
            resource="debts"
        )

    @retry_on_401_async
    async def create(self, flow_id: str, disbursement_date: Optional[str] = None, amount: Optional[dict] = None) -> str:
        if disbursement_date is not None:
            try:
                disbursement_date = dt.datetime.strptime(disbursement_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Invalid disbursement date, must be in the format YYYY-MM-DD")
        if amount is not None:
            amount = Money.parse_obj(amount)

        create_debt = CreateDebt.parse_obj({
            "amount": amount.dict(by_alias=True) if amount else None,
            "disbursed_at": disbursement_date.strftime("%Y-%m-%d") if disbursement_date else None,
            "flow_id": flow_id,
        })
        async with httpx.AsyncClient(base_url=self.altscore_client._cms_base_url) as client:
            response = await client.post(
                f"/{self.resource_version}/{self.resource}",
                headers=self.build_headers(),
                json=create_debt.dict(by_alias=True, exclude_none=True),
                timeout=30
            )
            raise_for_status_improved(response)
            response_json = response.json()
            return response_json.get("debtId")


class DebtsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(
            altscore_client=altscore_client,
            sync_resource=DebtSync,
            retrieve_data_model=DebtAPIDTO,
            create_data_model=None,
            update_data_model=None,
            resource="debts"
        )

    @retry_on_401
    def create(self, flow_id: str, disbursement_date: Optional[str] = None, amount: Optional[dict] = None) -> str:
        if disbursement_date is not None:
            try:
                disbursement_date = dt.datetime.strptime(disbursement_date, "%Y-%m-%d")
            except ValueError:
                raise ValueError("Invalid disbursement date, must be in the format YYYY-MM-DD")
        if amount is not None:
            amount = Money.parse_obj(amount)
        create_debt = CreateDebt.parse_obj({
            "amount": amount.dict(by_alias=True) if amount else None,
            "disbursed_at": disbursement_date.strftime("%Y-%m-%d") if disbursement_date else None,
            "flow_id": flow_id,
        })
        with httpx.Client(base_url=self.altscore_client._cms_base_url) as client:
            response = client.post(
                f"/{self.resource_version}/{self.resource}",
                headers=self.build_headers(),
                json=create_debt.dict(by_alias=True, exclude_none=True),
                timeout=30
            )
            raise_for_status_improved(response)
            response_json = response.json()
            return response_json.get("debtId")



import httpx
from typing import Optional, Dict, Any, Literal, Union
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.helpers import build_headers


HttpMethod = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]


class APIRequestSyncModule:
    """
    A module for making custom API requests to the Borrower Central API
    with automatic authentication headers.
    
    Example usage:
        # GET request with query params
        response = altscore.borrower_central.api.get(
            "/v1/borrowers",
            params={"page": 1, "per-page": 10}
        )
        
        # POST request with JSON body
        response = altscore.borrower_central.api.post(
            "/v1/borrowers",
            json={"persona": "individual", "label": "Test"}
        )
        
        # Generic request with any method
        response = altscore.borrower_central.api.request(
            method="PATCH",
            path="/v1/borrowers/123",
            json={"label": "Updated"}
        )
    """

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self, extra_headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """Build headers with auth + any extra headers provided."""
        headers = build_headers(self)
        if extra_headers:
            headers.update(extra_headers)
        return headers

    @property
    def base_url(self) -> str:
        return self.altscore_client._borrower_central_base_url

    @retry_on_401
    def request(
        self,
        method: HttpMethod,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """
        Make a custom API request to the Borrower Central API.
        
        Args:
            method: HTTP method (GET, POST, PUT, PATCH, DELETE)
            path: API path (e.g., "/v1/borrowers" or "/v1/borrowers/123")
            params: Query parameters
            json: JSON body (will be serialized)
            data: Form data body
            headers: Additional headers to merge with auth headers
            timeout: Request timeout in seconds (default: 30)
            raise_for_status: Whether to raise an exception on non-2xx responses (default: True)
            
        Returns:
            httpx.Response object
        """
        # Ensure path starts with /
        if not path.startswith("/"):
            path = f"/{path}"
        
        url = f"{self.base_url}{path}"
        merged_headers = self.build_headers(headers)
        
        with httpx.Client() as client:
            response = client.request(
                method=method,
                url=url,
                params=params,
                json=json,
                data=data,
                headers=merged_headers,
                timeout=timeout,
            )
            
            if raise_for_status:
                raise_for_status_improved(response)
            
            return response

    def get(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make a GET request."""
        return self.request(
            method="GET",
            path=path,
            params=params,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    def post(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make a POST request."""
        return self.request(
            method="POST",
            path=path,
            params=params,
            json=json,
            data=data,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    def put(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make a PUT request."""
        return self.request(
            method="PUT",
            path=path,
            params=params,
            json=json,
            data=data,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    def patch(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make a PATCH request."""
        return self.request(
            method="PATCH",
            path=path,
            params=params,
            json=json,
            data=data,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    def delete(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make a DELETE request."""
        return self.request(
            method="DELETE",
            path=path,
            params=params,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )


class APIRequestAsyncModule:
    """
    Async module for making custom API requests to the Borrower Central API
    with automatic authentication headers.
    
    Example usage:
        # GET request with query params
        response = await altscore.borrower_central.api.get(
            "/v1/borrowers",
            params={"page": 1, "per-page": 10}
        )
        
        # POST request with JSON body
        response = await altscore.borrower_central.api.post(
            "/v1/borrowers",
            json={"persona": "individual", "label": "Test"}
        )
        
        # Generic request with any method
        response = await altscore.borrower_central.api.request(
            method="PATCH",
            path="/v1/borrowers/123",
            json={"label": "Updated"}
        )
    """

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self, extra_headers: Optional[Dict[str, str]] = None) -> Dict[str, str]:
        """Build headers with auth + any extra headers provided."""
        headers = build_headers(self)
        if extra_headers:
            headers.update(extra_headers)
        return headers

    @property
    def base_url(self) -> str:
        return self.altscore_client._borrower_central_base_url

    @retry_on_401_async
    async def request(
        self,
        method: HttpMethod,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """
        Make a custom async API request to the Borrower Central API.
        
        Args:
            method: HTTP method (GET, POST, PUT, PATCH, DELETE)
            path: API path (e.g., "/v1/borrowers" or "/v1/borrowers/123")
            params: Query parameters
            json: JSON body (will be serialized)
            data: Form data body
            headers: Additional headers to merge with auth headers
            timeout: Request timeout in seconds (default: 30)
            raise_for_status: Whether to raise an exception on non-2xx responses (default: True)
            
        Returns:
            httpx.Response object
        """
        # Ensure path starts with /
        if not path.startswith("/"):
            path = f"/{path}"
        
        url = f"{self.base_url}{path}"
        merged_headers = self.build_headers(headers)
        
        async with httpx.AsyncClient() as client:
            response = await client.request(
                method=method,
                url=url,
                params=params,
                json=json,
                data=data,
                headers=merged_headers,
                timeout=timeout,
            )
            
            if raise_for_status:
                raise_for_status_improved(response)
            
            return response

    async def get(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make an async GET request."""
        return await self.request(
            method="GET",
            path=path,
            params=params,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    async def post(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make an async POST request."""
        return await self.request(
            method="POST",
            path=path,
            params=params,
            json=json,
            data=data,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    async def put(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make an async PUT request."""
        return await self.request(
            method="PUT",
            path=path,
            params=params,
            json=json,
            data=data,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    async def patch(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        json: Optional[Union[Dict[str, Any], list]] = None,
        data: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make an async PATCH request."""
        return await self.request(
            method="PATCH",
            path=path,
            params=params,
            json=json,
            data=data,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )

    async def delete(
        self,
        path: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 30,
        raise_for_status: bool = True,
    ) -> httpx.Response:
        """Make an async DELETE request."""
        return await self.request(
            method="DELETE",
            path=path,
            params=params,
            headers=headers,
            timeout=timeout,
            raise_for_status=raise_for_status,
        )



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


# DTO for deal contacts
class DealContactDTO(BaseModel):
    """Data transfer object for deal contacts"""
    id: str = Field(alias="id")
    deal_id: str = Field(alias="dealId")
    borrower_id: str = Field(alias="borrowerId")
    role_key: str = Field(alias="roleKey")
    is_primary: bool = Field(alias="isPrimary")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateDealContactRequest(BaseModel):
    """Model for creating a new deal contact"""
    deal_id: str = Field(alias="dealId")
    borrower_id: str = Field(alias="borrowerId")
    role_key: str = Field(alias="roleKey")
    is_primary: bool = Field(alias="isPrimary", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateDealContactRequest(BaseModel):
    """Model for updating a deal contact"""
    role_key: Optional[str] = Field(alias="roleKey", default=None)
    is_primary: Optional[bool] = Field(alias="isPrimary", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


# Resource classes for deal contacts
class DealContactSync(GenericSyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "deal-contacts", header_builder, renew_token, DealContactDTO.parse_obj(data))


class DealContactAsync(GenericAsyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "deal-contacts", header_builder, renew_token, DealContactDTO.parse_obj(data))


# Module for deal contacts - synchronous
class DealContactsSyncModule(GenericSyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=DealContactSync,
                         retrieve_data_model=DealContactDTO,
                         create_data_model=CreateDealContactRequest,
                         update_data_model=UpdateDealContactRequest,
                         resource="deal-contacts")

    @retry_on_401
    def get_by_deal_id(self, deal_id: str, page: int = 1, per_page: int = 100):
        """
        Get all contacts for a specific deal
        
        Args:
            deal_id: The ID of the deal
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of contacts for the deal
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/deal-contacts",
                params={
                    "deal-id": deal_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]

    @retry_on_401
    def get_by_borrower_id(self, borrower_id: str, page: int = 1, per_page: int = 100):
        """
        Get all deal contacts for a specific borrower
        
        Args:
            borrower_id: The ID of the borrower
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of deal contacts for the borrower
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/deal-contacts",
                params={
                    "borrower-id": borrower_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]

    @retry_on_401
    def get_by_deal_and_borrower(self, deal_id: str, borrower_id: str, page: int = 1, per_page: int = 100):
        """
        Get all contacts for a specific deal and borrower
        
        Args:
            deal_id: The ID of the deal
            borrower_id: The ID of the borrower
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of contacts for the deal and borrower
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/deal-contacts",
                params={
                    "deal-id": deal_id,
                    "borrower-id": borrower_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]
    
    @retry_on_401
    def get_by_role_key(self, role_key: str, page: int = 1, per_page: int = 100):
        """
        Get all contacts with a specific role
        
        Args:
            role_key: The role key to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of contacts with the specified role
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/deal-contacts",
                params={
                    "role-key": role_key,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]


# Module for deal contacts - asynchronous
class DealContactsAsyncModule(GenericAsyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=DealContactAsync,
                         retrieve_data_model=DealContactDTO,
                         create_data_model=CreateDealContactRequest,
                         update_data_model=UpdateDealContactRequest,
                         resource="deal-contacts")

    @retry_on_401_async
    async def get_by_deal_id(self, deal_id: str, page: int = 1, per_page: int = 100):
        """
        Get all contacts for a specific deal
        
        Args:
            deal_id: The ID of the deal
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of contacts for the deal
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/deal-contacts",
                params={
                    "deal-id": deal_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]

    @retry_on_401_async
    async def get_by_borrower_id(self, borrower_id: str, page: int = 1, per_page: int = 100):
        """
        Get all deal contacts for a specific borrower
        
        Args:
            borrower_id: The ID of the borrower
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of deal contacts for the borrower
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/deal-contacts",
                params={
                    "borrower-id": borrower_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]

    @retry_on_401_async
    async def get_by_deal_and_borrower(self, deal_id: str, borrower_id: str, page: int = 1, per_page: int = 100):
        """
        Get all contacts for a specific deal and borrower
        
        Args:
            deal_id: The ID of the deal
            borrower_id: The ID of the borrower
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of contacts for the deal and borrower
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/deal-contacts",
                params={
                    "deal-id": deal_id,
                    "borrower-id": borrower_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]
    
    @retry_on_401_async
    async def get_by_role_key(self, role_key: str, page: int = 1, per_page: int = 100):
        """
        Get all contacts with a specific role
        
        Args:
            role_key: The role key to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            List[DealContactDTO]: List of contacts with the specified role
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/deal-contacts",
                params={
                    "role-key": role_key,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return [DealContactDTO.parse_obj(data) for data in response.json()]



import httpx
from pydantic import BaseModel, Field
from typing import Optional, Dict, Any, List
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.helpers import build_headers


class PrintEntityRequest(BaseModel):
    """Request schema for printing an entity."""
    entity_type: str = Field(alias="entityType")
    entity_id: str = Field(alias="entityId")
    version: str = Field(alias="version", default="1")
    options: Optional[Dict[str, Any]] = Field(alias="options", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class EnrichedFieldValue(BaseModel):
    """Represents a field value enriched with data model metadata."""
    field_id: str = Field(alias="fieldId")
    key: str = Field(alias="key")
    label: Optional[str] = Field(alias="label", default=None)
    value: Any = Field(alias="value")
    data_type: str = Field(alias="dataType")
    tags: List[str] = Field(alias="tags", default=[])
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)
    allowed_values: Optional[List[Any]] = Field(alias="allowedValues", default=None)
    priority: Optional[int] = Field(alias="priority", default=None)
    order: Optional[int] = Field(alias="order", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class PrintEntityResponse(BaseModel):
    """Response schema for the print entity endpoint."""
    entity_type: str = Field(alias="entityType")
    entity_id: str = Field(alias="entityId")
    version: str = Field(alias="version")
    data: Dict[str, Any] = Field(alias="data")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

    def get_fields(self) -> List[EnrichedFieldValue]:
        """
        Parse and return the fields from the data as EnrichedFieldValue objects.
        Works for entity types that have a 'fields' list in their data.
        """
        fields_data = self.data.get("fields", [])
        return [EnrichedFieldValue.parse_obj(f) for f in fields_data]


class EntityPrintersSyncModule:
    """
    Module for printing entities in a format suitable for display/export.
    
    Example usage:
        # Print an asset
        result = altscore.borrower_central.entity_printers.print_entity(
            entity_type="asset",
            entity_id="asset-uuid-here"
        )
        print(result.data)  # The entity data
        
        # Get enriched fields
        fields = result.get_fields()
        for field in fields:
            print(f"{field.label}: {field.value}")
    """

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401
    def print_entity(
        self,
        entity_type: str,
        entity_id: str,
        version: str = "1",
        options: Optional[Dict[str, Any]] = None,
        timeout: int = 30
    ) -> PrintEntityResponse:
        """
        Print an entity's data in a format suitable for printing/export.
        
        Args:
            entity_type: The type of entity to print (e.g., "asset")
            entity_id: The ID of the entity to print
            version: The version of the print format (default: "1")
            options: Optional configuration options for printing
            timeout: Request timeout in seconds
            
        Returns:
            PrintEntityResponse containing the print-friendly entity data
        """
        url = f"{self.altscore_client._borrower_central_base_url}/v1/internal/entity-printers/print"
        
        request_data = PrintEntityRequest(
            entity_type=entity_type,
            entity_id=entity_id,
            version=version,
            options=options
        )
        
        with httpx.Client() as client:
            response = client.post(
                url,
                headers=self.build_headers(),
                json=request_data.dict(by_alias=True, exclude_none=True),
                timeout=timeout
            )
            raise_for_status_improved(response)
            return PrintEntityResponse.parse_obj(response.json())


class EntityPrintersAsyncModule:
    """
    Async module for printing entities in a format suitable for display/export.
    
    Example usage:
        # Print an asset
        result = await altscore.borrower_central.entity_printers.print_entity(
            entity_type="asset",
            entity_id="asset-uuid-here"
        )
        print(result.data)  # The entity data
        
        # Get enriched fields
        fields = result.get_fields()
        for field in fields:
            print(f"{field.label}: {field.value}")
    """

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def renew_token(self):
        self.altscore_client.renew_token()

    def build_headers(self):
        return build_headers(self)

    @retry_on_401_async
    async def print_entity(
        self,
        entity_type: str,
        entity_id: str,
        version: str = "1",
        options: Optional[Dict[str, Any]] = None,
        timeout: int = 30
    ) -> PrintEntityResponse:
        """
        Print an entity's data in a format suitable for printing/export.
        
        Args:
            entity_type: The type of entity to print (e.g., "asset")
            entity_id: The ID of the entity to print
            version: The version of the print format (default: "1")
            options: Optional configuration options for printing
            timeout: Request timeout in seconds
            
        Returns:
            PrintEntityResponse containing the print-friendly entity data
        """
        url = f"{self.altscore_client._borrower_central_base_url}/v1/internal/entity-printers/print"
        
        request_data = PrintEntityRequest(
            entity_type=entity_type,
            entity_id=entity_id,
            version=version,
            options=options
        )
        
        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self.build_headers(),
                json=request_data.dict(by_alias=True, exclude_none=True),
                timeout=timeout
            )
            raise_for_status_improved(response)
            return PrintEntityResponse.parse_obj(response.json())



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, List, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class IdentityAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: str = Field(alias="borrowerId")
    key: str = Field(alias="key")
    label: Optional[str] = Field(alias="label")
    value: Optional[str] = Field(alias="value")
    priority: Optional[int] = Field(alias="priority")
    tags: List[str] = Field(alias="tags")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")
    has_attachments: bool = Field(alias="hasAttachments", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateIdentityDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    key: str = Field(alias="key")
    value: Optional[str] = Field(alias="value")
    tags: List[str] = Field(alias="tags", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateIdentityDTO(BaseModel):
    value: Optional[str] = Field(alias="value", default=None)
    tags: Optional[List[str]] = Field(alias="tags", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class IdentitySync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "identities", header_builder, renew_token, IdentityAPIDTO.parse_obj(data))

    @retry_on_401
    def unmask(self) -> str:
        """
        Calls GET /v1/identities/{id}/unmask.
        Returns the plaintext value (requires bc.private.read and non-form_token principal).
        """
        with httpx.Client() as client:
            resp = client.get(
                f"{self.base_url}/v1/{self.resource}/{self.data.id}/unmask",
                headers=self._header_builder(),
                timeout=300,
            )
            raise_for_status_improved(resp)
            data = resp.json()
            return data.get("value")


class IdentityAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "identities", header_builder, renew_token, IdentityAPIDTO.parse_obj(data))

    @retry_on_401_async
    async def unmask(self) -> str:
        """
        Calls GET /v1/identities/{id}/unmask.
        Returns the plaintext value (requires bc.private.read and non-form_token principal).
        """
        async with httpx.AsyncClient() as client:
            resp = await client.get(
                f"{self.base_url}/v1/{self.resource}/{self.data.id}/unmask",
                headers=self._header_builder(),
                timeout=300,
            )
            raise_for_status_improved(resp)
            data = resp.json()
            return data.get("value")


class IdentitiesSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=IdentitySync,
                         retrieve_data_model=IdentityAPIDTO,
                         create_data_model=CreateIdentityDTO,
                         update_data_model=UpdateIdentityDTO,
                         resource="identities")

    @retry_on_401
    def find_by_key(self, key: str, borrower_id: str):
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            identities_found_req = client.get(
                "/v1/identities",
                params={
                    "key": key,
                    "borrower-id": borrower_id,
                    "per-page": 1,
                    "page": 1
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(identities_found_req)
            identities_found_data = identities_found_req.json()
            if len(identities_found_data) == 0:
                return None
            else:
                return self.retrieve(identities_found_data[0]["id"])

    @retry_on_401
    def unmask(self, identity_id: str) -> str:
        """
        Calls GET /v1/identities/{identity_id}/unmask.
        Returns the plaintext value (requires bc.private.read and non-form_token principal).
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            resp = client.get(
                f"/v1/identities/{identity_id}/unmask",
                headers=self.build_headers(),
                timeout=300,
            )
            raise_for_status_improved(resp)
            data = resp.json()
            return data.get("value")


class IdentitiesAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=IdentityAsync,
                         retrieve_data_model=IdentityAPIDTO,
                         create_data_model=CreateIdentityDTO,
                         update_data_model=UpdateIdentityDTO,
                         resource="identities")

    @retry_on_401_async
    async def find_by_key(self, key: str, borrower_id: str):
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            identities_found_req = await client.get(
                "/v1/identities",
                params={
                    "key": key,
                    "borrower-id": borrower_id,
                    "per-page": 1,
                    "page": 1
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(identities_found_req)
            identities_found_data = identities_found_req.json()
            if len(identities_found_data) == 0:
                return None
            else:
                return await self.retrieve(identities_found_data[0]["id"])

    @retry_on_401_async
    async def unmask(self, identity_id: str) -> str:
        """
        Calls GET /v1/identities/{identity_id}/unmask.
        Returns the plaintext value (requires bc.private.read and non-form_token principal).
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            resp = await client.get(
                f"/v1/identities/{identity_id}/unmask",
                headers=self.build_headers(),
                timeout=300,
            )
            raise_for_status_improved(resp)
            data = resp.json()
            return data.get("value")



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class MessageAPIDTO(BaseModel):
    """DTO for message API responses"""
    id: str = Field(alias="id")
    borrower_id: str = Field(alias="borrowerId")
    channel: str = Field(alias="channel")
    channel_customer_id: str = Field(alias="channelCustomerId")
    conversation_id: str = Field(alias="conversationId")
    content: str = Field(alias="content")
    sender_type: str = Field(alias="senderType")
    sender_id: Optional[str] = Field(alias="senderId")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata")
    created_at: str = Field(alias="createdAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class MessageCreate(BaseModel):
    channel: str = Field(alias="channel")
    conversation_id: str = Field(alias="conversationId")
    borrower_id: str = Field(alias="borrowerId")
    content: str = Field(alias="content")
    channel_customer_id: str = Field(alias="channelCustomerId")
    sender_type: str = Field(alias="senderType")
    sender_id: Optional[str] = Field(alias="senderId", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class MessageSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/messages", header_builder, renew_token,
                         MessageAPIDTO.parse_obj(data))


class MessageAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/messages", header_builder, renew_token,
                         MessageAPIDTO.parse_obj(data))


class MessageSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=MessageSync, retrieve_data_model=MessageAPIDTO,
                         create_data_model=MessageCreate, update_data_model=None,
                         resource="/conversational/messages")


class MessageAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=MessageSync, retrieve_data_model=MessageAPIDTO,
                         create_data_model=MessageCreate, update_data_model=None,
                         resource="/conversational/messages")



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class DataModelAPIDTO(BaseModel):
    id: str = Field(alias="id")
    path: str = Field(alias="path")
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    entity_type: str = Field(alias="entityType")
    priority: Optional[int] = Field(alias="priority", default=None)
    order: Optional[int] = Field(alias="order", default=None)
    allowed_values: Optional[List[Any]] = Field(alias="allowedValues", default=None)
    data_type: Optional[str] = Field(alias="dataType", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata")
    is_segmentation_field: Optional[bool] = Field(alias="isSegmentationField", default=False)
    is_sensitive: Optional[bool] = Field(alias="isSensitive", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class DataModelCreate(BaseModel):
    path: Optional[str] = Field(alias="path", default="")
    key: str = Field(alias="key")
    label: str = Field(alias="label")
    entity_type: str = Field(alias="entityType")
    priority: Optional[int] = Field(alias="priority", default=None)
    allowed_values: Optional[List[Any]] = Field(alias="allowedValues", default=None)
    order: Optional[int] = Field(alias="order", default=None)
    data_type: Optional[str] = Field(alias="dataType", default=None)
    metadata: Optional[dict] = Field(alias="metadata", default={})
    is_segmentation_field: Optional[bool] = Field(alias="isSegmentationField", default=False)
    is_sensitive: Optional[bool] = Field(alias="isSensitive", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class DataModelUpdate(BaseModel):
    path: Optional[str] = Field(alias="path", default="")
    key: Optional[str] = Field(alias="key")
    label: Optional[str] = Field(alias="label")
    priority: Optional[int] = Field(alias="priority", default=None)
    order: Optional[int] = Field(alias="order", default=None)
    allowed_values: Optional[List[Any]] = Field(alias="allowedValues", default=None)
    data_type: Optional[str] = Field(alias="dataType", default=None)
    metadata: Optional[dict] = Field(alias="metadata", default={})
    is_segmentation_field: Optional[bool] = Field(alias="isSegmentationField", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class DataModelSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "data-models", header_builder, renew_token, DataModelAPIDTO.parse_obj(data))


class DataModelAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "data-models", header_builder, renew_token, DataModelAPIDTO.parse_obj(data))


class DataModelSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=DataModelSync, retrieve_data_model=DataModelAPIDTO,
                         create_data_model=DataModelCreate, update_data_model=DataModelUpdate, resource="data-models")

    @retry_on_401
    def make_sensitive(self, data_model_id: str):
        """
        Calls PUT /v1/data-models/{data_model_id}/make-sensitive.
        Returns the updated DataModel resource.
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            resp = client.put(
                f"/v1/data-models/{data_model_id}/make-sensitive",
                headers=self.build_headers(),
                timeout=300,
            )
            raise_for_status_improved(resp)
            return self.sync_resource(
                base_url=self.altscore_client._borrower_central_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=self.retrieve_data_model.parse_obj(resp.json()),
            )


class DataModelAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=DataModelSync, retrieve_data_model=DataModelAPIDTO,
                         create_data_model=DataModelCreate, update_data_model=DataModelUpdate, resource="data-models")

    @retry_on_401_async
    async def make_sensitive(self, data_model_id: str):
        """
        Calls PUT /v1/data-models/{data_model_id}/make-sensitive.
        Returns the updated DataModel resource.
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            resp = await client.put(
                f"/v1/data-models/{data_model_id}/make-sensitive",
                headers=self.build_headers(),
                timeout=300,
            )
            raise_for_status_improved(resp)
            return self.async_resource(
                base_url=self.altscore_client._borrower_central_base_url,
                header_builder=self.build_headers,
                renew_token=self.renew_token,
                data=self.retrieve_data_model.parse_obj(resp.json()),
            )



from pydantic import BaseModel, Field
from typing import Optional, List, Dict, Any
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class ConversationAPIDTO(BaseModel):
    """DTO for conversation API responses"""
    id: str = Field(alias="id")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    connector_id: str = Field(alias="connectorId")
    connector_sender_id: Optional[str] = Field(alias="connectorSenderId", default=None)
    channel_user_id: str = Field(alias="channelUserId")
    channel: str = Field(alias="channel")
    state: str = Field(alias="state")
    current_human_agent_id: Optional[str] = Field(alias="currentHumanAgentId", default=None)
    current_ai_agent_id: Optional[str] = Field(alias="currentAIAgentId", default=None)
    bot_state: Optional[Dict[str, Any]] = Field(alias="botState", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)
    pending_messages_count: Optional[int] = Field(alias="messagePendingCount", default=0)
    status_chat: Optional[str] = Field(alias="statusChat", default=None)
    last_message_at: Optional[str] = Field(alias="lastMessageAt", default=None)
    notices: Optional[List[Dict[str, Any]]] = Field(alias="notices", default=[])
    is_hidden: Optional[bool] = Field(alias="isHidden", default=False)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ConversationCreate(BaseModel):
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    connector_id: str = Field(alias="connectorId")
    connector_sender_id: str = Field(alias="connectorSenderId")
    channel_user_id: str = Field(alias="channelUserId")
    channel: Optional[str] = Field(alias="channelId", default="whatsapp")
    state: Optional[str] = Field(alias="state", default=None)
    bot_state: Optional[Dict[str, Any]] = Field(alias="botState", default=None)
    current_human_agent_id: Optional[str] = Field(alias="currentHumanAgentId", default=None)
    current_ai_agent_id: Optional[str] = Field(alias="currentAIAgentId", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ConversationUpdate(BaseModel):
    state: Optional[str] = Field(alias="state")
    current_human_agent_id: Optional[str] = Field(alias="currentHumanAgentId")  # Fixed alias to be consistent
    current_ai_agent_id: Optional[str] = Field(alias="currentAIAgentId")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ConversationSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/conversations", header_builder, renew_token,
                         ConversationAPIDTO.parse_obj(data))


class ConversationAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "/conversational/conversations", header_builder, renew_token,
                         ConversationAPIDTO.parse_obj(data))


class ConversationSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, sync_resource=ConversationSync, retrieve_data_model=ConversationAPIDTO,
                         create_data_model=ConversationCreate, update_data_model=ConversationUpdate,
                         resource="/conversational/conversations")


class ConversationAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client, async_resource=ConversationAsync, retrieve_data_model=ConversationAPIDTO,
                         create_data_model=ConversationCreate, update_data_model=ConversationUpdate,
                         resource="/conversational/conversations")




from typing import List, Optional
from mimetypes import guess_type

from pydantic import BaseModel, EmailStr

class Email(BaseModel):
    email: EmailStr

class Attachment(BaseModel):
    type: str
    filename: str
    content: str

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class MailBody(BaseModel):
    to: List[Email]
    cc: Optional[List[Email]] = []
    bcc: Optional[List[Email]] = []
    subject: str
    content: str
    attachments: List[Attachment] = []
    from_name: Optional[str]

    def add_attachment(self, filename: str, content: str, file_type = None):
        attachment_type = file_type if file_type is not None else guess_type(filename)[0]
        if attachment_type is None:
            raise Exception("No attachment type specified nor could be guessed")

        self.attachments.append(
            Attachment(type=attachment_type, filename=filename, content=content)
        )

    def to_dict(self):
        return {
            'to': [m.dict() for m in self.to],
            'cc': [m.dict() for m in self.cc],
            'bcc': [m.dict() for m in self.bcc],
            'subject': self.subject,
            'content': self.content,
            'attachments': [a.dict(by_alias=True) for a in self.attachments],
            'from_name': self.from_name,
        }


from typing import Optional, List
from pydantic import BaseModel, Field


class CalendarHoliday(BaseModel):
    day: int = Field(alias="day")
    month: str = Field(alias="month")


class DPACalendarAPIDTO(BaseModel):
    id: str = Field(alias="calendarId")
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    exclusionDays: Optional[List[str]] = Field(alias="exclusionDays", default=None)
    holidays: Optional[List[CalendarHoliday]] = Field(alias="holidays", default=None)
    type: str = Field(alias="type")
    year: int = Field(alias="year")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class CreateCalendarAPIDTO(BaseModel):
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    exclusionDays: Optional[List[str]] = Field(alias="exclusionDays", default=None)
    holidays: Optional[List[str]] = Field(alias="holidays", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True



from pydantic import BaseModel, Field
from typing import Optional, Any, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from altscore.borrower_central.model.bank_accounts import Money


class BankAccountBalanceAPIDTO(BaseModel):
    """
    Bank account balance information at a specific point in time
    """
    id: str = Field(alias="id", description="Unique balance record identifier")
    bank_account_id: str = Field(alias="bankAccountId", description="Associated bank account ID")
    current_balance: Money = Field(alias="currentBalance", description="Current available balance")
    available_balance: Optional[Money] = Field(alias="availableBalance", default=None, description="Available for transactions")
    ledger_balance: Optional[Money] = Field(alias="ledgerBalance", default=None, description="Book balance including pending")
    balance_date: str = Field(alias="balanceDate", description="Date this balance is effective")
    effective_date: Optional[str] = Field(alias="effectiveDate", default=None, description="When balance becomes effective")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None, description="Additional custom data")
    created_at: str = Field(alias="createdAt", description="Balance record creation timestamp")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None, description="Last update timestamp")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateBankAccountBalanceDTO(BaseModel):
    """
    Request payload for creating a new bank account balance record
    """
    bank_account_id: str = Field(alias="bankAccountId")
    current_balance: Money = Field(alias="currentBalance")
    available_balance: Optional[Money] = Field(alias="availableBalance", default=None)
    ledger_balance: Optional[Money] = Field(alias="ledgerBalance", default=None)
    balance_date: str = Field(alias="balanceDate", description="Date in YYYY-MM-DD format")
    effective_date: Optional[str] = Field(alias="effectiveDate", default=None, description="Date in YYYY-MM-DD format")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateBankAccountBalanceDTO(BaseModel):
    """
    Request payload for updating an existing bank account balance record
    """
    current_balance: Optional[Money] = Field(alias="currentBalance", default=None)
    available_balance: Optional[Money] = Field(alias="availableBalance", default=None)
    ledger_balance: Optional[Money] = Field(alias="ledgerBalance", default=None)
    balance_date: Optional[str] = Field(alias="balanceDate", default=None, description="Date in YYYY-MM-DD format")
    effective_date: Optional[str] = Field(alias="effectiveDate", default=None, description="Date in YYYY-MM-DD format")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BankAccountBalanceSync(GenericSyncResource):
    """
    Synchronous resource for bank account balance operations
    """
    
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "banking/account-balances", header_builder,
                         renew_token, BankAccountBalanceAPIDTO.parse_obj(data))


class BankAccountBalanceAsync(GenericAsyncResource):
    """
    Asynchronous resource for bank account balance operations
    """
    
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "banking/account-balances", header_builder,
                         renew_token, BankAccountBalanceAPIDTO.parse_obj(data))


class BankAccountBalancesSyncModule(GenericSyncModule):
    """
    Synchronous module for bank account balance management
    """
    
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=BankAccountBalanceSync,
                         retrieve_data_model=BankAccountBalanceAPIDTO,
                         create_data_model=CreateBankAccountBalanceDTO,
                         update_data_model=UpdateBankAccountBalanceDTO,
                         resource="banking/account-balances")

class BankAccountBalancesAsyncModule(GenericAsyncModule):
    """
    Asynchronous module for bank account balance management
    """
    
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=BankAccountBalanceAsync,
                         retrieve_data_model=BankAccountBalanceAPIDTO,
                         create_data_model=CreateBankAccountBalanceDTO,
                         update_data_model=UpdateBankAccountBalanceDTO,
                         resource="banking/account-balances")


import httpx
from pydantic import BaseModel, Field
from typing import Dict, Any, List, Optional
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from altscore.borrower_central.helpers import build_headers


class ValidationErrorLocation(BaseModel):
    row: Optional[int] = Field(alias="row", default=None)
    column: Optional[str] = Field(alias="column", default=None)
    field: Optional[str] = Field(alias="field", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ValidationError(BaseModel):
    location: ValidationErrorLocation = Field(alias="location")
    value: Any = Field(alias="value")
    message: str = Field(alias="message")
    error_type: Optional[str] = Field(alias="errorType", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ValidationResult(BaseModel):
    success: bool = Field(alias="success")
    errors: List[ValidationError] = Field(alias="errors", default=[])

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class InputValidationSyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def build_headers(self):
        return build_headers(self)

    @retry_on_401
    def validate_single(
        self,
        data: Dict[str, Any],
        schema: Dict[str, Any],
        language: str = "en"
    ) -> ValidationResult:
        """
        Validate a single JSON object against a JSON schema.

        Args:
            data: The JSON object to validate
            schema: JSON Schema definition
            language: Language for error messages (default: "en")

        Returns:
            ValidationResult with success status and any validation errors
        """
        url = f"{self.altscore_client._borrower_central_base_url}/v1/input-validation/"

        payload = {
            "data": data,
            "schema": schema,
            "language": language
        }

        with httpx.Client() as client:
            response = client.post(
                url,
                headers=self.build_headers(),
                json=payload,
                timeout=30
            )
            raise_for_status_improved(response)
            return ValidationResult.parse_obj(response.json())


class InputValidationAsyncModule:

    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def build_headers(self):
        return build_headers(self)

    @retry_on_401_async
    async def validate_single(
        self,
        data: Dict[str, Any],
        schema: Dict[str, Any],
        language: str = "en"
    ) -> ValidationResult:
        """
        Validate a single JSON object against a JSON schema.

        Args:
            data: The JSON object to validate
            schema: JSON Schema definition
            language: Language for error messages (default: "en")

        Returns:
            ValidationResult with success status and any validation errors
        """
        url = f"{self.altscore_client._borrower_central_base_url}/v1/input-validation/"

        payload = {
            "data": data,
            "schema": schema,
            "language": language
        }

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self.build_headers(),
                json=payload,
                timeout=30
            )
            raise_for_status_improved(response)
            return ValidationResult.parse_obj(response.json())



from altscore.cms.model.partners import PartnersSyncModule, PartnersAsyncModule
from altscore.cms.model.clients import ClientsSyncModule, ClientsAsyncModule
from altscore.cms.model.dpas import DPAFlowsAsyncModule, DPAFlowsSyncModule
from altscore.cms.model.debts import DebtsAsyncModule, DebtsSyncModule
from altscore.cms.model.disbursements import DisbursementSyncModule, DisbursementAsyncModule
from altscore.cms.model.payment_orders import PaymentOrdersAsyncModule, PaymentOrdersSyncModule
from altscore.cms.helpers import build_headers


class CMSSync:

    def __init__(self, altscore_client):
        self.clients = ClientsSyncModule(altscore_client)
        self.partners = PartnersSyncModule(altscore_client)
        self.dpas = DPAFlowsSyncModule(altscore_client)
        self.debts = DebtsSyncModule(altscore_client)
        self.payment_orders = PaymentOrdersSyncModule(altscore_client)
        self.disbursements = DisbursementSyncModule(altscore_client)


class CMSAsync:

    def __init__(self, altscore_client):
        self.clients = ClientsAsyncModule(altscore_client)
        self.partners = PartnersAsyncModule(altscore_client)
        self.dpas = DPAFlowsAsyncModule(altscore_client)
        self.debts = DebtsAsyncModule(altscore_client)
        self.payment_orders = PaymentOrdersAsyncModule(altscore_client)
        self.disbursements = DisbursementAsyncModule(altscore_client)



from pydantic import BaseModel, Field
from typing import Optional, Any, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from altscore.borrower_central.model.bank_accounts import Money


class TransactionDates(BaseModel):
    """
    Date information for bank transactions
    """
    transaction_date: str = Field(
        alias="transactionDate",
        description="Date when the transaction occurred"
    )
    posting_date: Optional[str] = Field(
        default=None,
        alias="postingDate", 
        description="Date when transaction was posted to account"
    )


    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BankTransactionAPIDTO(BaseModel):
    """
    Bank transaction record with comprehensive transaction details
    """
    id: str = Field(alias="id", description="Unique transaction identifier")
    bank_account_id: str = Field(alias="bankAccountId", description="Associated bank account ID")
    transaction_type: str = Field(alias="transactionType", description="Type of transaction: debit, credit, transfer, etc.")
    amount: Money = Field(alias="amount", description="Transaction amount")
    running_balance: Optional[Money] = Field(alias="runningBalance", default=None, description="Account balance after transaction")
    description: Optional[str] = Field(alias="description", default=None, description="Transaction description")
    reference_number: Optional[str] = Field(alias="referenceNumber", default=None, description="Bank reference number")
    counterparty_name: Optional[str] = Field(alias="counterpartyName", default=None, description="Other party in transaction")
    counterparty_account: Optional[str] = Field(alias="counterpartyAccount", default=None, description="Other party's account")
    transaction_dates: TransactionDates = Field(alias="transactionDates", description="Transaction date information")
    parsed_raw: Optional[Dict[str, Any]] = Field(alias="parsedRaw", default=None, description="Parsed raw transaction data")
    confidence_score: Optional[float] = Field(alias="confidenceScore", default=None, description="Parsing confidence score")
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None, description="Additional custom data")
    created_at: str = Field(alias="createdAt", description="Transaction record creation timestamp")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None, description="Last update timestamp")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateBankTransactionDTO(BaseModel):
    """
    Request payload for creating a new bank transaction record
    """
    bank_account_id: str = Field(alias="bankAccountId")
    transaction_type: str = Field(alias="transactionType")
    amount: Money = Field(alias="amount")
    running_balance: Optional[Money] = Field(alias="runningBalance", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    reference_number: Optional[str] = Field(alias="referenceNumber", default=None)
    counterparty_name: Optional[str] = Field(alias="counterpartyName", default=None)
    counterparty_account: Optional[str] = Field(alias="counterpartyAccount", default=None)
    transaction_dates: TransactionDates = Field(alias="transactionDates")
    parsed_raw: Optional[Dict[str, Any]] = Field(alias="parsedRaw", default=None)
    confidence_score: Optional[float] = Field(alias="confidenceScore", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateBankTransactionDTO(BaseModel):
    """
    Request payload for updating an existing bank transaction record
    """
    transaction_type: Optional[str] = Field(alias="transactionType", default=None)
    amount: Optional[Money] = Field(alias="amount", default=None)
    running_balance: Optional[Money] = Field(alias="runningBalance", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    counterparty_name: Optional[str] = Field(alias="counterpartyName", default=None)
    counterparty_account: Optional[str] = Field(alias="counterpartyAccount", default=None)
    transaction_dates: Optional[TransactionDates] = Field(alias="transactionDates", default=None)
    parsed_raw: Optional[Dict[str, Any]] = Field(alias="parsedRaw", default=None)
    confidence_score: Optional[float] = Field(alias="confidenceScore", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class BankTransactionSync(GenericSyncResource):
    """
    Synchronous resource for bank transaction operations
    """
    
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "banking/transactions", header_builder,
                         renew_token, BankTransactionAPIDTO.parse_obj(data))


class BankTransactionAsync(GenericAsyncResource):
    """
    Asynchronous resource for bank transaction operations
    """
    
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "banking/transactions", header_builder,
                         renew_token, BankTransactionAPIDTO.parse_obj(data))


class BankTransactionsSyncModule(GenericSyncModule):
    """
    Synchronous module for bank transaction management
    """
    
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=BankTransactionSync,
                         retrieve_data_model=BankTransactionAPIDTO,
                         create_data_model=CreateBankTransactionDTO,
                         update_data_model=UpdateBankTransactionDTO,
                         resource="banking/transactions")

class BankTransactionsAsyncModule(GenericAsyncModule):
    """
    Asynchronous module for bank transaction management
    """
    
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=BankTransactionAsync,
                         retrieve_data_model=BankTransactionAPIDTO,
                         create_data_model=CreateBankTransactionDTO,
                         update_data_model=UpdateBankTransactionDTO,
                         resource="banking/transactions")


from uuid import uuid4
import datetime as dt

from fuzzywuzzy import process

from altscore.borrower_central.model.borrower import BorrowerSync, BorrowerAsync
from altscore.borrower_central.model.store_packages import altdata_source_slug
from altscore.altdata.model.data_request import SourceConfig
from altscore.macros.validate_inputs import validate_borrower_data
from typing import Optional, Tuple, List, Dict, Union
import asyncio
from loguru import logger


class MacrosSync:
    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    def create_borrower(self, borrower_data: dict) -> Optional[BorrowerSync]:
        identities_to_create, borrower_fields_to_create = validate_borrower_data(borrower_data)
        borrower_id = self.altscore_client.borrower_central.borrowers.create(
            {
                "label": borrower_data.get("label"),
                "persona": borrower_data["persona"],
                "tags": borrower_data.get("tags", []),
                "flag": borrower_data.get("flag"),
                "riskRating": borrower_data.get("risk_rating") or borrower_data.get("riskRating"),
                "repaymentRiskRating": borrower_data.get("repayment_risk_rating") or \
                                       borrower_data.get("repaymentRiskRating"),
            }
        )
        try:
            for identity_key in identities_to_create:
                key = identity_key.split(".")[-1]
                value = borrower_data[identity_key]
                self.altscore_client.borrower_central.identities.create(
                    {
                        "borrower_id": borrower_id,
                        "key": key,
                        "value": value
                    }
                )
            for field_key in borrower_fields_to_create:
                key = field_key.split(".")[-1]
                value = borrower_data[field_key]
                self.altscore_client.borrower_central.borrower_fields.create(
                    {
                        "borrower_id": borrower_id,
                        "key": key,
                        "value": value
                    }
                )

            if len(borrower_data.get("points_of_contact", [])) > 0:
                for point_of_contact in borrower_data.get("points_of_contact"):
                    self.altscore_client.borrower_central.points_of_contact.create(
                        {
                            "borrower_id": borrower_id,
                            **point_of_contact
                        }
                    )

            if len(borrower_data.get("identities", [])) > 0:
                for identity in borrower_data["identities"]:
                    identity_id = self.altscore_client.borrower_central.identities.create(
                        {
                            "borrower_id": borrower_id,
                            **identity
                        }
                    )
                    if len(identity.get("attachments", [])) > 0:
                        identity_obj = self.altscore_client.borrower_central.identities.retieve(identity_id)
                        for attachment in identity["attachments"]:
                            identity_obj.post_attachment(attachment)

            if len(borrower_data.get("borrower_fields", [])) > 0:
                for borrower_field in borrower_data["borrower_fields"]:
                    self.altscore_client.borrower_central.borrower_fields.create(
                        {
                            "borrower_id": borrower_id,
                            **borrower_field
                        }
                    )

            if len(borrower_data.get("addresses", [])) > 0:
                for address in borrower_data["addresses"]:
                    address_id = self.altscore_client.borrower_central.addresses.create(
                        {
                            "borrower_id": borrower_id,
                            **address
                        }
                    )
                    if len(address.get("attachments", [])) > 0:
                        address_obj = self.altscore_client.borrower_central.addresses.retieve(address_id)
                        for attachment in address["attachments"]:
                            address_obj.post_attachment(attachment)

            if len(borrower_data.get("documents", [])) > 0:
                for document in borrower_data.get("documents"):
                    document_id = self.altscore_client.borrower_central.documents.create(
                        {
                            "borrower_id": borrower_id,
                            **document
                        }
                    )
                    if len(document.get("attachments", [])) > 0:
                        document_obj = self.altscore_client.borrower_central.documents.retieve(document_id)
                        for attachment in document["attachments"]:
                            document_obj.post_attachment(attachment)

            return self.altscore_client.borrower_central.borrowers.retrieve(borrower_id)
        except Exception as e:
            logger.error(f"Error creating borrower, deleting {borrower_id}")
            logger.exception(e)
            self.altscore_client.borrower_central.borrowers.delete(borrower_id)
            return None

    def new_cms_client_from_borrower(
            self, borrower_id: str,
            partner_id: str = None,
            legal_name_identity_key: Optional[str] = None,
            tax_id_identity_key: Optional[str] = None,
            external_id_identity_key: str = None,
            dba_identity_key: Optional[str] = None
    ) -> str:

        def find_identity_value_or_error(_borrower, identity_key):
            identity = _borrower.get_identity_by_key(key=identity_key)
            if identity is None:
                raise LookupError(f"Identity {identity_key} not found for borrower {borrower_id}")
            else:
                return identity.data.value

        borrower = self.altscore_client.borrower_central.borrowers.retrieve(borrower_id)
        if borrower is None:
            raise LookupError(f"Borrower {borrower_id} not found")

        if external_id_identity_key is not None:
            external_id = find_identity_value_or_error(borrower, external_id_identity_key)
        else:
            # carefull as the system validates unique external_id per partner per tenant
            external_id = borrower_id

        if legal_name_identity_key is not None:
            legal_name = find_identity_value_or_error(borrower, legal_name_identity_key)
        else:
            legal_name = "N/A"

        if borrower.data.persona == "business":
            if dba_identity_key is not None:
                dba = find_identity_value_or_error(borrower, dba_identity_key)
            else:
                dba = legal_name
        else:
            dba = "N/A"

        if tax_id_identity_key is not None:
            tax_id = find_identity_value_or_error(borrower, tax_id_identity_key)
        else:
            # carefull as the system validates unique tax_id per partner per tenant
            tax_id = f"NA-{str(uuid4())[:8]}"

        address = borrower.get_main_address()
        if address is None:
            address = "N/A"
        else:
            address = address.data.get_address_str()

        email = borrower.get_main_point_of_contact(contact_method="email")
        if email is None:
            email = "na@na.com"
        else:
            email = email.data.value

        phone = borrower.get_main_point_of_contact(contact_method="phone")
        if phone is None:
            phone = "N/A"
        else:
            phone = phone.data.value

        client_data = {"externalId": external_id, "legalName": legal_name, "taxId": tax_id, "dba": dba,
                       "address": address, "emailAddress": email, "phoneNumber": phone, "partnerId": partner_id,
                       "borrowerId": borrower_id}

        # see if there is already a cms client with the given external id
        cms_client = self.altscore_client.cms.clients.retrieve_by_external_id(external_id=external_id)
        if cms_client is not None:
            # associate the borrower with the client
            borrower.associate_cms_client_id(cms_client.data.id)
            # update the client with the new legal name and tax id
            self.altscore_client.cms.clients.patch(
                resource_id=cms_client.data.id,
                patch_data={
                    "legalName": legal_name,
                    "taxId": tax_id,
                    "emailAddress": email,
                    "borrowerId": borrower_id
                }
            )
            return cms_client.data.id
        else:
            client_id = self.altscore_client.cms.clients.create(new_entity_data=client_data)
            borrower.associate_cms_client_id(client_id)
            return client_id

    def get_unique_borrower_field_values(self, field_key: str):
        field_values = self.altscore_client.borrower_central.borrower_fields.count_distinct_values(field_key)
        return [item["value"] for item in field_values]

    def evaluate_value_migration(self, target_values: list, current_values: list, threshold = 85):
        approved_changes, doubt = [], []
        for current in current_values:
            new_val = process.extractOne(current, target_values, score_cutoff=threshold)
            if new_val is not None and current != new_val[0]: # don't add if is the same value
                approved_changes.append((current, new_val[0]))
            elif new_val is None:
                doubt.append(current)

        return approved_changes, doubt

    def migrate_borrower_field_allowed_values(
            self,
            field_key: str,
            replace_values: List[Tuple[str, str]],
            allowed_values= None
    ):
        data_model = self.altscore_client.borrower_central.data_models.query(
            entity_type="borrower_field",
            key=field_key,
        )

        if data_model is None or data_model == []:
            raise Exception("No data model found with key {}".format(field_key))

        for value, target in replace_values:
            self.altscore_client.borrower_central.borrower_fields.bulk_update_field_values(field_key, value, target)

        if allowed_values is not None:
            self.altscore_client.borrower_central.data_models.patch(
                data_model[0].data.id,
                {
                    "allowedValues": allowed_values
                }
            )

    def enrich_borrower(
            self,
            borrower_id: str,
            sources: List[Dict],
            input_keys: Dict,
            data_age_minutes: int = 360,
            timeout_seconds: int = 120,
    ) -> dict:
        """
        Full AltData enrichment cycle with caching.

        Checks freshness for each source, calls AltData for stale/missing ones,
        stores results as packages. Returns a summary dict.

        Args:
            borrower_id: The borrower to enrich
            sources: List of source configs, e.g. [{"sourceId": "ECU-PUB-0002", "version": "v1"}]
            input_keys: Identity keys for AltData lookup, e.g. {"person_id": "123"}
            data_age_minutes: Max age in minutes before a package is considered stale (default 360)
            timeout_seconds: AltData sync call timeout (default 120)

        Returns:
            dict with keys: source_results, all_sources_ok, sources_created, sources_fresh, sources_failed
        """
        bc = self.altscore_client.borrower_central
        ad = self.altscore_client.altdata

        data_age = dt.timedelta(minutes=data_age_minutes)
        source_results = []
        sources_to_call = []

        for src in sources:
            s = SourceConfig.parse_obj(src) if isinstance(src, dict) else src
            slug = altdata_source_slug(s.source_id, s.version)
            pkg = bc.store_packages.retrieve_source_package(
                source_id=slug, borrower_id=borrower_id, data_age=data_age
            )
            if pkg is not None:
                source_results.append({
                    "source_slug": slug, "package_id": pkg.data.id, "status": "fresh"
                })
            else:
                sources_to_call.append(s)

        if sources_to_call:
            try:
                result = ad.requests.new_sync(
                    input_keys=input_keys,
                    sources_config=sources_to_call,
                    timeout=timeout_seconds,
                )
                for call in result.call_summary:
                    slug = altdata_source_slug(call.source_id, call.version)
                    if call.is_success:
                        pkg_id = bc.store_packages.create_from_altdata_request_result(
                            borrower_id=borrower_id,
                            source_id=call.source_id,
                            altdata_request_result=result,
                        )
                        source_results.append({
                            "source_slug": slug, "package_id": pkg_id, "status": "created"
                        })
                    else:
                        source_results.append({
                            "source_slug": slug, "status": "failed",
                            "error": call.error_message or "Unknown error"
                        })
            except Exception as e:
                for s in sources_to_call:
                    slug = altdata_source_slug(s.source_id, s.version)
                    if not any(r["source_slug"] == slug for r in source_results):
                        source_results.append({
                            "source_slug": slug, "status": "failed", "error": str(e)
                        })

        created = sum(1 for r in source_results if r["status"] == "created")
        fresh = sum(1 for r in source_results if r["status"] == "fresh")
        failed = sum(1 for r in source_results if r["status"] == "failed")

        return {
            "source_results": source_results,
            "all_sources_ok": failed == 0,
            "sources_created": created,
            "sources_fresh": fresh,
            "sources_failed": failed,
        }

    def find_or_create_borrower(
            self,
            identity_key: str,
            identity_value: str,
            persona: str = "individual",
            label: Optional[str] = None,
    ) -> dict:
        """
        Idempotent borrower lookup/creation by identity key.

        Searches for an existing borrower with the given identity key+value.
        If found, returns the borrower_id. If not, creates a new borrower
        with that identity attached.

        Args:
            identity_key: Identity key to search by (e.g. "person_id", "tax_id", "email")
            identity_value: The value to match
            persona: "individual" or "company" (used only on creation)
            label: Display name (defaults to identity_value if not provided)

        Returns:
            dict with keys: borrower_id, created
        """
        bc = self.altscore_client.borrower_central
        identities = bc.identities.query(key=identity_key, value=identity_value, per_page=1)
        if identities:
            return {"borrower_id": identities[0].data.borrower_id, "created": False}

        borrower_id = bc.borrowers.create({
            "persona": persona,
            "label": label or identity_value,
        })
        bc.identities.create({
            "borrowerId": borrower_id,
            "key": identity_key,
            "value": identity_value,
        })
        return {"borrower_id": borrower_id, "created": True}

    def evaluate(
            self,
            evaluator_alias: str,
            evaluator_version: str,
            reference_id: str,
            data: Dict,
            entities: Optional[List[Dict]] = None,
            execution_id: Optional[str] = None,
    ) -> dict:
        """
        Run an evaluator by alias/version with a simplified interface.

        Builds the EvaluatorInput internally (instance + entities), calls the
        evaluator, and returns the output as a plain dict.

        Args:
            evaluator_alias: Evaluator alias (e.g. "scoring", "kyc")
            evaluator_version: Evaluator version (e.g. "v2", "v3")
            reference_id: Subject identifier -- typically the borrower_id
            data: Dict of variables for the evaluator instance
            entities: Optional list of entity dicts (co-debtors, guarantors).
                      Each must have entityId, role, data. Defaults to [].
            execution_id: Optional workflow execution ID. When provided, stored
                          in the instance data as _execution_id so the evaluation
                          is traceable back to the execution that triggered it.

        Returns:
            dict with keys: score, scorecard, metrics, rules, decision.
            Raises Exception if the evaluator returns an error.
        """
        bc = self.altscore_client.borrower_central

        instance_data = dict(data)
        if execution_id is not None:
            instance_data["_execution_id"] = execution_id

        evaluator_input = {
            "instance": {
                "referenceId": reference_id,
                "referenceDate": dt.datetime.now().isoformat(),
                "data": instance_data,
            },
            "entities": entities or [],
        }

        result = bc.evaluators.evaluate(
            evaluator_input=evaluator_input,
            evaluator_alias=evaluator_alias,
            evaluator_version=evaluator_version,
        )

        if hasattr(result, "traceback"):
            raise Exception(f"Evaluator error: {result.detail}")

        return result.dict(by_alias=True)

    def get_borrower_metrics(
            self,
            borrower_id: str,
            metric_keys: List[str],
            default=-999999,
            none_on_sentinel: Optional[List[str]] = None,
    ) -> dict:
        """
        Batch-extract borrower metrics with sentinel handling.

        Retrieves the borrower once, then loops through metric_keys calling
        get_metric_by_key() for each. Returns a flat dict.

        Args:
            borrower_id: The borrower to query
            metric_keys: List of metric key strings to extract
            default: Default value for missing metrics (default: -999999)
            none_on_sentinel: List of keys where the sentinel value should be
                              replaced with None instead of kept as-is. Useful for
                              keys like "days_since_first_sale" where -999999 means
                              "not applicable" rather than a valid number.

        Returns:
            dict mapping each metric key to its value
        """
        bc = self.altscore_client.borrower_central
        none_keys = set(none_on_sentinel or [])

        borrower = bc.borrowers.retrieve(borrower_id)
        if borrower is None:
            raise ValueError(f"Borrower {borrower_id} not found")

        metrics = {}
        for key in metric_keys:
            metric_obj = borrower.get_metric_by_key(key)
            if metric_obj is not None:
                value = metric_obj.data.value
                if value == default and key in none_keys:
                    metrics[key] = None
                else:
                    metrics[key] = value
            else:
                metrics[key] = None if key in none_keys else default
        return metrics

    def create_alerts_from_rules(
            self,
            borrower_id: str,
            rules: List[Dict],
            execution_id: Optional[str] = None,
            level_mapping: Optional[Dict[str, int]] = None,
            default_level: int = 0,
    ) -> List[Dict]:
        """
        Create borrower alerts from evaluator rule results.

        Filters rules where hit == True, derives alert level from a prefix
        mapping, and creates alerts via the API. Duplicates (HTTP 409) are
        skipped silently; all other errors propagate.

        Args:
            borrower_id: The borrower to create alerts for
            rules: The "rules" list from evaluator output
            execution_id: Optional execution ID to tie alerts to the workflow run
            level_mapping: Dict mapping rule code prefixes to alert levels.
                           E.g. {"DR_D": 2, "DR_R": 1, "DR_AP": 2}.
                           The first matching prefix wins.
            default_level: Alert level when no prefix matches (default: 0)

        Returns:
            List of alert body dicts that were created (or attempted)
        """
        from httpx import HTTPStatusError

        bc = self.altscore_client.borrower_central
        level_map = level_mapping or {}

        hit_rules = [r for r in rules if r.get("hit") is True]
        alerts = []

        for rule in hit_rules:
            code = rule.get("code", "")
            level = default_level
            for prefix, lvl in level_map.items():
                if code.startswith(prefix):
                    level = lvl
                    break

            alert_body = {
                "borrowerId": borrower_id,
                "ruleCode": code.replace("_", "-"),
                "level": level,
                "message": rule.get("label") or rule.get("value") or code,
            }
            if execution_id is not None:
                alert_body["referenceId"] = execution_id

            try:
                bc.alerts.create(alert_body)
            except HTTPStatusError as e:
                if e.response.status_code == 409:
                    pass  # duplicate alert, expected
                else:
                    raise
            alerts.append(alert_body)

        return alerts


class MacrosAsync:
    def __init__(self, altscore_client):
        self.altscore_client = altscore_client

    async def create_borrower(self, borrower_data: dict) -> Optional[BorrowerAsync]:
        identities_to_create, borrower_fields_to_create = validate_borrower_data(borrower_data)
        borrower_id = await self.altscore_client.borrower_central.borrowers.create(
            {
                "label": borrower_data.get("label"),
                "persona": borrower_data["persona"],
                "tags": borrower_data.get("tags", []),
                "flag": borrower_data.get("flag"),
                "riskRating": borrower_data.get("risk_rating") or borrower_data.get("riskRating"),
                "repaymentRiskRating": borrower_data.get("repayment_risk_rating") or \
                                       borrower_data.get("repaymentRiskRating"),
            }
        )
        try:
            calls = []
            for identity_key in identities_to_create:
                key = identity_key.split(".")[-1]
                value = borrower_data[identity_key]
                calls.append(self.altscore_client.borrower_central.identities.create(
                    {
                        "borrower_id": borrower_id,
                        "key": key,
                        "value": value
                    }
                ))
            for field_key in borrower_fields_to_create:
                key = field_key.split(".")[-1]
                value = borrower_data[field_key]
                calls.append(self.altscore_client.borrower_central.borrower_fields.create(
                    {
                        "borrower_id": borrower_id,
                        "key": key,
                        "value": value
                    }
                ))
            if len(borrower_data.get("points_of_contact", [])) > 0:
                for point_of_contact in borrower_data.get("points_of_contact"):
                    calls.append(self.altscore_client.borrower_central.points_of_contact.create(
                        {
                            "borrower_id": borrower_id,
                            **point_of_contact
                        }
                    ))
            await asyncio.gather(*calls)

            # if this entities have attachments we cannot make them concurrently
            if len(borrower_data.get("identities", [])) > 0:
                for identity in borrower_data["identities"]:
                    identity_id = await self.altscore_client.borrower_central.identities.create(
                        {
                            "borrower_id": borrower_id,
                            **identity
                        }
                    )
                    if len(identity.get("attachments", [])) > 0:
                        identity_obj = await self.altscore_client.borrower_central.identities.retieve(identity_id)
                        for attachment in identity["attachments"]:
                            await identity_obj.post_attachment(attachment)

            if len(borrower_data.get("borrower_fields", [])) > 0:
                for borrower_field in borrower_data["borrower_fields"]:
                    await self.altscore_client.borrower_central.borrower_fields.create(
                        {
                            "borrower_id": borrower_id,
                            **borrower_field
                        }
                    )

            if len(borrower_data.get("addresses", [])) > 0:
                for address in borrower_data["addresses"]:
                    address_id = await self.altscore_client.borrower_central.addresses.create(
                        {
                            "borrower_id": borrower_id,
                            **address
                        }
                    )
                    if len(address.get("attachments", [])) > 0:
                        address_obj = await self.altscore_client.borrower_central.addresses.retieve(address_id)
                        for attachment in address["attachments"]:
                            await address_obj.post_attachment(attachment)

            if len(borrower_data.get("documents", [])) > 0:
                for document in borrower_data.get("documents"):
                    document_id = await self.altscore_client.borrower_central.documents.create(
                        {
                            "borrower_id": borrower_id,
                            **document
                        }
                    )
                    if len(document.get("attachments", [])) > 0:
                        document_obj = await self.altscore_client.borrower_central.documents.retieve(document_id)
                        for attachment in document["attachments"]:
                            await document_obj.post_attachment(attachment)

            return await self.altscore_client.borrower_central.borrowers.retrieve(borrower_id)
        except Exception as e:
            logger.error(f"Error creating borrower, deleting {borrower_id}")
            logger.exception(e)
            await self.altscore_client.borrower_central.borrowers.delete(borrower_id)
            return None

    async def new_cms_client_from_borrower(
            self, borrower_id: str, partner_id: str, legal_name_identity_key: str, tax_id_identity_key: str,
            external_id_identity_key: str = None, dba_identity_key: Optional[str] = None
    ) -> str:
        async def find_identity_value_or_error(_borrower, identity_key):
            identity = await _borrower.get_identity_by_key(key=identity_key)
            if identity is None:
                raise LookupError(f"Identity {identity_key} not found for borrower {borrower_id}")
            else:
                return identity.data.value

        borrower = await self.altscore_client.borrower_central.borrowers.retrieve(borrower_id)
        if borrower is None:
            raise LookupError(f"Borrower {borrower_id} not found")

        if external_id_identity_key is not None:
            external_id = await find_identity_value_or_error(borrower, external_id_identity_key)
        else:
            # carefull as the system validates unique external_id per partner per tenant
            external_id = borrower_id

        if legal_name_identity_key is not None:
            legal_name = await find_identity_value_or_error(borrower, legal_name_identity_key)
        else:
            legal_name = "N/A"

        if borrower.data.persona == "business":
            if dba_identity_key is not None:
                dba = await find_identity_value_or_error(borrower, dba_identity_key)
            else:
                dba = legal_name
        else:
            dba = "N/A"

        if tax_id_identity_key is not None:
            tax_id = await find_identity_value_or_error(borrower, tax_id_identity_key)
        else:
            # carefull as the system validates unique tax_id per partner per tenant
            tax_id = f"NA-{str(uuid4())[:8]}"

        address = await borrower.get_main_address()
        if address is None:
            address = "N/A"
        else:
            address = address.data.get_address_str()

        email = await borrower.get_main_point_of_contact(contact_method="email")
        if email is None:
            email = "na@na.com"
        else:
            email = email.data.value

        phone = await borrower.get_main_point_of_contact(contact_method="phone")
        if phone is None:
            phone = "N/A"
        else:
            phone = phone.data.value

        client_data = {"externalId": external_id, "legalName": legal_name, "taxId": tax_id, "dba": dba,
                       "address": address, "emailAddress": email, "phoneNumber": phone, "partnerId": partner_id,
                       "borrowerId": borrower_id}
        # see if there is already a cms client with the given external id
        cms_client = await self.altscore_client.cms.clients.retrieve_by_external_id(external_id=external_id)
        if cms_client is not None:
            # associate the borrower with the client
            await borrower.associate_cms_client_id(cms_client.data.id)
            # update the client with the new legal name and tax id
            await self.altscore_client.cms.clients.patch(
                resource_id=cms_client.data.id,
                patch_data={
                    "legalName": legal_name,
                    "taxId": tax_id,
                    "emailAddress": email,
                    "borrowerId": borrower_id
                }
            )
            return cms_client.data.id
        else:
            client_id = await self.altscore_client.cms.clients.create(new_entity_data=client_data)
            await borrower.associate_cms_client_id(client_id)
            return client_id

    async def get_unique_borrower_field_values(self, field_key: str):
        field_values = await self.altscore_client.borrower_central.borrower_fields.count_distinct_values(field_key)
        return [item["value"] for item in field_values]

    async def evaluate_value_migration(self, target_values: list, current_values: list, threshold = 85):
        approved_changes, doubt = [], []
        for current in current_values:
            new_val = process.extractOne(current, target_values, score_cutoff=threshold)
            if new_val is not None and current != new_val[0]: # don't add if is the same value
                approved_changes.append((current, new_val[0]))
            elif new_val is None:
                doubt.append(current)

        return approved_changes, doubt

    async def migrate_borrower_field_allowed_values(
            self,
            field_key: str,
            replace_values: List[Tuple[str, str]],
            new_allowed_values: List[str]
    ):
        replacing_values = [val[1] for val in replace_values]
        if not set(replacing_values).issubset(set(new_allowed_values)):
            raise ValueError("Trying to set values not present in the new allowed values")

        data_model = await self.altscore_client.borrower_central.data_models.query(
            entity_type="borrower_field",
            key=field_key,
        )

        if data_model is None or data_model == []:
            raise Exception("No data model found with key {}".format(field_key))

        async_calls = []

        for value, target in replace_values:
            call = asyncio.create_task(self.altscore_client.borrower_central.borrower_fields.bulk_update_field_values(
                field_key, value, target
            ))
            async_calls.append(call)

        await asyncio.gather(*async_calls)

        await self.altscore_client.borrower_central.data_models.patch(
            data_model[0].data.id,
            {
                "allowedValues": new_allowed_values
            }
        )

    async def enrich_borrower(
            self,
            borrower_id: str,
            sources: List[Dict],
            input_keys: Dict,
            data_age_minutes: int = 360,
            timeout_seconds: int = 120,
    ) -> dict:
        """
        Full AltData enrichment cycle with caching.

        Checks freshness for each source, calls AltData for stale/missing ones,
        stores results as packages. Returns a summary dict.

        Args:
            borrower_id: The borrower to enrich
            sources: List of source configs, e.g. [{"sourceId": "ECU-PUB-0002", "version": "v1"}]
            input_keys: Identity keys for AltData lookup, e.g. {"person_id": "123"}
            data_age_minutes: Max age in minutes before a package is considered stale (default 360)
            timeout_seconds: AltData sync call timeout (default 120)

        Returns:
            dict with keys: source_results, all_sources_ok, sources_created, sources_fresh, sources_failed
        """
        bc = self.altscore_client.borrower_central
        ad = self.altscore_client.altdata

        data_age = dt.timedelta(minutes=data_age_minutes)
        source_results = []
        sources_to_call = []

        for src in sources:
            s = SourceConfig.parse_obj(src) if isinstance(src, dict) else src
            slug = altdata_source_slug(s.source_id, s.version)
            pkg = await bc.store_packages.retrieve_source_package(
                source_id=slug, borrower_id=borrower_id, data_age=data_age
            )
            if pkg is not None:
                source_results.append({
                    "source_slug": slug, "package_id": pkg.data.id, "status": "fresh"
                })
            else:
                sources_to_call.append(s)

        if sources_to_call:
            try:
                result = await ad.requests.new_sync(
                    input_keys=input_keys,
                    sources_config=sources_to_call,
                    timeout=timeout_seconds,
                )
                for call in result.call_summary:
                    slug = altdata_source_slug(call.source_id, call.version)
                    if call.is_success:
                        pkg_id = await bc.store_packages.create_from_altdata_request_result(
                            borrower_id=borrower_id,
                            source_id=call.source_id,
                            altdata_request_result=result,
                        )
                        source_results.append({
                            "source_slug": slug, "package_id": pkg_id, "status": "created"
                        })
                    else:
                        source_results.append({
                            "source_slug": slug, "status": "failed",
                            "error": call.error_message or "Unknown error"
                        })
            except Exception as e:
                for s in sources_to_call:
                    slug = altdata_source_slug(s.source_id, s.version)
                    if not any(r["source_slug"] == slug for r in source_results):
                        source_results.append({
                            "source_slug": slug, "status": "failed", "error": str(e)
                        })

        created = sum(1 for r in source_results if r["status"] == "created")
        fresh = sum(1 for r in source_results if r["status"] == "fresh")
        failed = sum(1 for r in source_results if r["status"] == "failed")

        return {
            "source_results": source_results,
            "all_sources_ok": failed == 0,
            "sources_created": created,
            "sources_fresh": fresh,
            "sources_failed": failed,
        }

    async def find_or_create_borrower(
            self,
            identity_key: str,
            identity_value: str,
            persona: str = "individual",
            label: Optional[str] = None,
    ) -> dict:
        """
        Idempotent borrower lookup/creation by identity key.

        Searches for an existing borrower with the given identity key+value.
        If found, returns the borrower_id. If not, creates a new borrower
        with that identity attached.

        Args:
            identity_key: Identity key to search by (e.g. "person_id", "tax_id", "email")
            identity_value: The value to match
            persona: "individual" or "company" (used only on creation)
            label: Display name (defaults to identity_value if not provided)

        Returns:
            dict with keys: borrower_id, created
        """
        bc = self.altscore_client.borrower_central
        identities = await bc.identities.query(key=identity_key, value=identity_value, per_page=1)
        if identities:
            return {"borrower_id": identities[0].data.borrower_id, "created": False}

        borrower_id = await bc.borrowers.create({
            "persona": persona,
            "label": label or identity_value,
        })
        await bc.identities.create({
            "borrowerId": borrower_id,
            "key": identity_key,
            "value": identity_value,
        })
        return {"borrower_id": borrower_id, "created": True}

    async def evaluate(
            self,
            evaluator_alias: str,
            evaluator_version: str,
            reference_id: str,
            data: Dict,
            entities: Optional[List[Dict]] = None,
            execution_id: Optional[str] = None,
    ) -> dict:
        """
        Run an evaluator by alias/version with a simplified interface.

        Builds the EvaluatorInput internally (instance + entities), calls the
        evaluator, and returns the output as a plain dict.

        Args:
            evaluator_alias: Evaluator alias (e.g. "scoring", "kyc")
            evaluator_version: Evaluator version (e.g. "v2", "v3")
            reference_id: Subject identifier -- typically the borrower_id
            data: Dict of variables for the evaluator instance
            entities: Optional list of entity dicts (co-debtors, guarantors).
                      Each must have entityId, role, data. Defaults to [].
            execution_id: Optional workflow execution ID. When provided, stored
                          in the instance data as _execution_id so the evaluation
                          is traceable back to the execution that triggered it.

        Returns:
            dict with keys: score, scorecard, metrics, rules, decision.
            Raises Exception if the evaluator returns an error.
        """
        bc = self.altscore_client.borrower_central

        instance_data = dict(data)
        if execution_id is not None:
            instance_data["_execution_id"] = execution_id

        evaluator_input = {
            "instance": {
                "referenceId": reference_id,
                "referenceDate": dt.datetime.now().isoformat(),
                "data": instance_data,
            },
            "entities": entities or [],
        }

        result = await bc.evaluators.evaluate(
            evaluator_input=evaluator_input,
            evaluator_alias=evaluator_alias,
            evaluator_version=evaluator_version,
        )

        if hasattr(result, "traceback"):
            raise Exception(f"Evaluator error: {result.detail}")

        return result.dict(by_alias=True)

    async def get_borrower_metrics(
            self,
            borrower_id: str,
            metric_keys: List[str],
            default=-999999,
            none_on_sentinel: Optional[List[str]] = None,
    ) -> dict:
        """
        Batch-extract borrower metrics with sentinel handling.

        Retrieves the borrower once, then loops through metric_keys calling
        get_metric_by_key() for each. Returns a flat dict.

        Args:
            borrower_id: The borrower to query
            metric_keys: List of metric key strings to extract
            default: Default value for missing metrics (default: -999999)
            none_on_sentinel: List of keys where the sentinel value should be
                              replaced with None instead of kept as-is.

        Returns:
            dict mapping each metric key to its value
        """
        bc = self.altscore_client.borrower_central
        none_keys = set(none_on_sentinel or [])

        borrower = await bc.borrowers.retrieve(borrower_id)
        if borrower is None:
            raise ValueError(f"Borrower {borrower_id} not found")

        metrics = {}
        for key in metric_keys:
            metric_obj = await borrower.get_metric_by_key(key)
            if metric_obj is not None:
                value = metric_obj.data.value
                if value == default and key in none_keys:
                    metrics[key] = None
                else:
                    metrics[key] = value
            else:
                metrics[key] = None if key in none_keys else default
        return metrics

    async def create_alerts_from_rules(
            self,
            borrower_id: str,
            rules: List[Dict],
            execution_id: Optional[str] = None,
            level_mapping: Optional[Dict[str, int]] = None,
            default_level: int = 0,
    ) -> List[Dict]:
        """
        Create borrower alerts from evaluator rule results.

        Filters rules where hit == True, derives alert level from a prefix
        mapping, and creates alerts via the API. Duplicates (HTTP 409) are
        skipped silently; all other errors propagate.

        Args:
            borrower_id: The borrower to create alerts for
            rules: The "rules" list from evaluator output
            execution_id: Optional execution ID to tie alerts to the workflow run
            level_mapping: Dict mapping rule code prefixes to alert levels.
                           E.g. {"DR_D": 2, "DR_R": 1, "DR_AP": 2}.
            default_level: Alert level when no prefix matches (default: 0)

        Returns:
            List of alert body dicts that were created (or attempted)
        """
        from httpx import HTTPStatusError

        bc = self.altscore_client.borrower_central
        level_map = level_mapping or {}

        hit_rules = [r for r in rules if r.get("hit") is True]
        alerts = []

        for rule in hit_rules:
            code = rule.get("code", "")
            level = default_level
            for prefix, lvl in level_map.items():
                if code.startswith(prefix):
                    level = lvl
                    break

            alert_body = {
                "borrowerId": borrower_id,
                "ruleCode": code.replace("_", "-"),
                "level": level,
                "message": rule.get("label") or rule.get("value") or code,
            }
            if execution_id is not None:
                alert_body["referenceId"] = execution_id

            try:
                await bc.alerts.create(alert_body)
            except HTTPStatusError as e:
                if e.response.status_code == 409:
                    pass  # duplicate alert, expected
                else:
                    raise
            alerts.append(alert_body)

        return alerts



from altscore.borrower_central.model.analytics import AnalyticsAsyncModule, AnalyticsSyncModule
from altscore.borrower_central.model.borrower import BorrowersAsyncModule, BorrowersSyncModule
from altscore.borrower_central.model.category import CategoryAsyncModule, CategorySyncModule
from altscore.borrower_central.model.conversational_templates import WhatsAppTemplateSync, WhatsAppTemplateSyncModule, \
    WhatsAppTemplateAsyncModule
from altscore.borrower_central.model.data_models import DataModelAsyncModule, DataModelSyncModule
from altscore.borrower_central.model.executions import ExecutionAsyncModule, ExecutionSyncModule
from altscore.borrower_central.model.tools import ReportGeneratorAsyncModule, ReportGeneratorSyncModule, \
    CommunicationsAsyncModule, CommunicationsSyncModule
from altscore.borrower_central.model.form_templates import FormTemplatesAsyncModule, FormTemplatesSyncModule
from altscore.borrower_central.model.otp_templates import OTPTemplatesAsyncModule, OTPTemplatesSyncModule
from altscore.borrower_central.model.addresses import AddressesAsyncModule, AddressesSyncModule
from altscore.borrower_central.model.authorizations import AuthorizationsAsyncModule, AuthorizationsSyncModule
from altscore.borrower_central.model.borrower_fields import BorrowerFieldsAsyncModule, BorrowerFieldsSyncModule
from altscore.borrower_central.model.documents import DocumentsAsyncModule, DocumentsSyncModule
from altscore.borrower_central.model.identities import IdentitiesAsyncModule, IdentitiesSyncModule
from altscore.borrower_central.model.points_of_contact import PointOfContactAsyncModule, PointOfContactSyncModule
from altscore.borrower_central.model.relationships import RelationshipsAsyncModule, RelationshipsSyncModule
from altscore.borrower_central.model.store_packages import PackagesAsyncModule, PackagesSyncModule
from altscore.borrower_central.model.store_sources import SourcesAsyncModule, SourcesSyncModule
from altscore.borrower_central.model.store_secrets import SecretsAsyncModule, SecretsSyncModule
from altscore.borrower_central.model.workflows import WorkflowsAsyncModule, WorkflowsSyncModule
from altscore.borrower_central.model.evaluators import EvaluatorAsyncModule, EvaluatorSyncModule
from altscore.borrower_central.model.forms import FormsAsyncModule, FormsSyncModule
from altscore.borrower_central.model.policy_alerts import AlertsAsyncModule, AlertsSyncModule
from altscore.borrower_central.model.policy_rules import RulesAsyncModule, RulesSyncModule
from altscore.borrower_central.model.integrations.sat import SatIntegrationAsyncModule, SatIntegrationSyncModule
from altscore.borrower_central.model.integrations.sri import SriIntegrationAsyncModule, SriIntegrationSyncModule
from altscore.borrower_central.model.automations import AutomationsAsyncModule, AutomationsSyncModule
from altscore.borrower_central.model.stages import StagesAsyncModule, StagesSyncModule
from altscore.borrower_central.model.risk_ratings import RiskRatingsAsyncModule, RiskRatingsSyncModule
from altscore.borrower_central.model.usecases import UsecasesAsyncModule, UsecasesSyncModule
from altscore.borrower_central.model.metrics import MetricsAsyncModule, MetricsSyncModule
from altscore.borrower_central.model.policy_policies import PolicyAsyncModule, PolicySyncModule
from altscore.borrower_central.model.kpis import KpisAsyncModule, KpisSyncModule
from altscore.borrower_central.model.steps import StepsAsyncModule, StepsSyncModule
from altscore.borrower_central.model.list_of_similar import ListOfSimilarAsyncModule, ListOfSimilarSyncModule
from altscore.borrower_central.model.execution_batches import ExecutionBatchAsyncModule, ExecutionBatchSyncModule
from altscore.borrower_central.model.cms_settings import CMSSettingsAsyncModule, CMSSettingsSyncModule
from altscore.borrower_central.model.conversational_channel_connectors import ChannelConnectorAsyncModule, \
    ChannelConnectorSyncModule
from altscore.borrower_central.model.conversational_conversations import ConversationAsyncModule, ConversationSyncModule
from altscore.borrower_central.model.conversational_messages import MessageAsyncModule, MessageSyncModule
from altscore.borrower_central.model.custom_reports import CustomReportsSyncModule, CustomReportsAsyncModule
from altscore.borrower_central.model.report_templates import ReportTemplateAsyncModule, \
    ReportTemplateSyncModule
from altscore.borrower_central.model.deals import DealsAsyncModule, DealsSyncModule
from altscore.borrower_central.model.deal_fields import DealFieldsAsyncModule, DealFieldsSyncModule
from altscore.borrower_central.model.assets import AssetsAsyncModule, AssetsSyncModule
from altscore.borrower_central.model.asset_fields import AssetFieldsAsyncModule, AssetFieldsSyncModule
from altscore.borrower_central.model.deal_steps import DealStepsAsyncModule, DealStepsSyncModule
from altscore.borrower_central.model.deal_contacts import DealContactsAsyncModule, DealContactsSyncModule
from altscore.borrower_central.model.accounting_documents import AccountingDocumentsAsyncModule, \
    AccountingDocumentsSyncModule
from altscore.borrower_central.model.accounting_line_items import LineItemsAsyncModule, LineItemsSyncModule
from altscore.borrower_central.model.bank_accounts import BankAccountsSyncModule, \
    BankAccountsAsyncModule
from altscore.borrower_central.model.bank_account_balances import BankAccountBalancesSyncModule, \
    BankAccountBalancesAsyncModule
from altscore.borrower_central.model.bank_transactions import BankTransactionsSyncModule, \
    BankTransactionsAsyncModule
from altscore.borrower_central.model.actionables import ActionablesSyncModule, ActionablesAsyncModule
from altscore.borrower_central.model.input_validation import InputValidationSyncModule, InputValidationAsyncModule
from altscore.borrower_central.model.api_request import APIRequestSyncModule, APIRequestAsyncModule
from altscore.borrower_central.model.entity_printers import EntityPrintersSyncModule, EntityPrintersAsyncModule


class BorrowerCentralAsync:
    def __init__(self, altscore_client):
        self.addresses = AddressesAsyncModule(altscore_client)
        self.authorizations = AuthorizationsAsyncModule(altscore_client)
        self.automations = AutomationsAsyncModule(altscore_client)
        self.analytics = AnalyticsAsyncModule(altscore_client)
        self.borrowers = BorrowersAsyncModule(altscore_client)
        self.borrower_fields = BorrowerFieldsAsyncModule(altscore_client)
        self.metrics = MetricsAsyncModule(altscore_client)
        self.data_models = DataModelAsyncModule(altscore_client)
        self.documents = DocumentsAsyncModule(altscore_client)
        self.stages = StagesAsyncModule(altscore_client)
        self.risk_ratings = RiskRatingsAsyncModule(altscore_client)
        self.executions = ExecutionAsyncModule(altscore_client)
        self.identities = IdentitiesAsyncModule(altscore_client)
        self.workflows = WorkflowsAsyncModule(altscore_client)
        self.points_of_contact = PointOfContactAsyncModule(altscore_client)
        self.relationships = RelationshipsAsyncModule(altscore_client)
        self.report_generator = ReportGeneratorAsyncModule(altscore_client)
        self.form_templates = FormTemplatesAsyncModule(altscore_client)
        self.otp_templates = OTPTemplatesAsyncModule(altscore_client)
        self.store_packages = PackagesAsyncModule(altscore_client)
        self.store_sources = SourcesAsyncModule(altscore_client)
        self.store_secrets = SecretsAsyncModule(altscore_client)
        self.evaluators = EvaluatorAsyncModule(altscore_client)
        self.forms = FormsAsyncModule(altscore_client)
        self.alerts = AlertsAsyncModule(altscore_client)
        self.rules = RulesAsyncModule(altscore_client)
        self.policies = PolicyAsyncModule(altscore_client)
        self.sat_integration = SatIntegrationAsyncModule(altscore_client)
        self.sri_integration = SriIntegrationAsyncModule(altscore_client)
        self.usecases = UsecasesAsyncModule(altscore_client)
        self.kpis = KpisAsyncModule(altscore_client)
        self.steps = StepsAsyncModule(altscore_client)
        self.list_of_similar = ListOfSimilarAsyncModule(altscore_client)
        self.execution_batches = ExecutionBatchAsyncModule(altscore_client)
        self.categories = CategoryAsyncModule(altscore_client)
        self.cms_settings = CMSSettingsAsyncModule(altscore_client)
        self.communications = CommunicationsAsyncModule(altscore_client)
        self.conversational_channel_connectors = ChannelConnectorAsyncModule(altscore_client)
        self.conversational_conversations = ConversationAsyncModule(altscore_client)
        self.conversational_messages = MessageAsyncModule(altscore_client)
        self.custom_reports = ReportTemplateAsyncModule(altscore_client)
        self.report_templates = ReportTemplateAsyncModule(altscore_client)
        self.deals = DealsAsyncModule(altscore_client)
        self.deal_fields = DealFieldsAsyncModule(altscore_client)
        self.deal_steps = DealStepsAsyncModule(altscore_client)
        self.assets = AssetsAsyncModule(altscore_client)
        self.asset_fields = AssetFieldsAsyncModule(altscore_client)
        self.deal_contacts = DealContactsAsyncModule(altscore_client)
        self.accounting_documents = AccountingDocumentsAsyncModule(altscore_client)
        self.accounting_line_items = LineItemsAsyncModule(altscore_client)
        self.bank_accounts = BankAccountsAsyncModule(altscore_client)
        self.bank_account_balances = BankAccountBalancesAsyncModule(altscore_client)
        self.bank_transactions = BankTransactionsAsyncModule(altscore_client)
        self.conversational_templates = WhatsAppTemplateAsyncModule(altscore_client)
        self.actionables = ActionablesAsyncModule(altscore_client)
        self.input_validation = InputValidationAsyncModule(altscore_client)
        self.api = APIRequestAsyncModule(altscore_client)
        self.entity_printers = EntityPrintersAsyncModule(altscore_client)


class BorrowerCentralSync:
    def __init__(self, altscore_client):
        self.addresses = AddressesSyncModule(altscore_client)
        self.authorizations = AuthorizationsSyncModule(altscore_client)
        self.automations = AutomationsSyncModule(altscore_client)
        self.analytics = AnalyticsSyncModule(altscore_client)
        self.borrowers = BorrowersSyncModule(altscore_client)
        self.borrower_fields = BorrowerFieldsSyncModule(altscore_client)
        self.metrics = MetricsSyncModule(altscore_client)
        self.data_models = DataModelSyncModule(altscore_client)
        self.documents = DocumentsSyncModule(altscore_client)
        self.stages = StagesSyncModule(altscore_client)
        self.risk_ratings = RiskRatingsSyncModule(altscore_client)
        self.executions = ExecutionSyncModule(altscore_client)
        self.workflows = WorkflowsSyncModule(altscore_client)
        self.identities = IdentitiesSyncModule(altscore_client)
        self.points_of_contact = PointOfContactSyncModule(altscore_client)
        self.relationships = RelationshipsSyncModule(altscore_client)
        self.report_generator = ReportGeneratorSyncModule(altscore_client)
        self.form_templates = FormTemplatesSyncModule(altscore_client)
        self.otp_templates = OTPTemplatesSyncModule(altscore_client)
        self.store_packages = PackagesSyncModule(altscore_client)
        self.store_sources = SourcesSyncModule(altscore_client)
        self.store_secrets = SecretsSyncModule(altscore_client)
        self.evaluators = EvaluatorSyncModule(altscore_client)
        self.forms = FormsSyncModule(altscore_client)
        self.alerts = AlertsSyncModule(altscore_client)
        self.rules = RulesSyncModule(altscore_client)
        self.policies = PolicySyncModule(altscore_client)
        self.sat_integration = SatIntegrationSyncModule(altscore_client)
        self.sri_integration = SriIntegrationSyncModule(altscore_client)
        self.usecases = UsecasesSyncModule(altscore_client)
        self.kpis = KpisSyncModule(altscore_client)
        self.steps = StepsSyncModule(altscore_client)
        self.list_of_similar = ListOfSimilarSyncModule(altscore_client)
        self.execution_batches = ExecutionBatchSyncModule(altscore_client)
        self.categories = CategorySyncModule(altscore_client)
        self.cms_settings = CMSSettingsSyncModule(altscore_client)
        self.communications = CommunicationsSyncModule(altscore_client)
        self.conversational_channel_connectors = ChannelConnectorSyncModule(altscore_client)
        self.conversational_conversations = ConversationSyncModule(altscore_client)
        self.conversational_messages = MessageSyncModule(altscore_client)
        self.custom_reports = CustomReportsSyncModule(altscore_client)
        self.report_templates = ReportTemplateSyncModule(altscore_client)
        self.deals = DealsSyncModule(altscore_client)
        self.deal_fields = DealFieldsSyncModule(altscore_client)
        self.deal_steps = DealStepsSyncModule(altscore_client)
        self.assets = AssetsSyncModule(altscore_client)
        self.asset_fields = AssetFieldsSyncModule(altscore_client)
        self.deal_contacts = DealContactsSyncModule(altscore_client)
        self.accounting_documents = AccountingDocumentsSyncModule(altscore_client)
        self.accounting_line_items = LineItemsSyncModule(altscore_client)
        self.bank_accounts = BankAccountsSyncModule(altscore_client)
        self.bank_account_balances = BankAccountBalancesSyncModule(altscore_client)
        self.bank_transactions = BankTransactionsSyncModule(altscore_client)
        self.conversational_templates = WhatsAppTemplateSyncModule(altscore_client)
        self.actionables = ActionablesSyncModule(altscore_client)
        self.input_validation = InputValidationSyncModule(altscore_client)
        self.api = APIRequestSyncModule(altscore_client)
        self.entity_printers = EntityPrintersSyncModule(altscore_client)


from pydantic import BaseModel, Field
from typing import Optional, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


class RelationshipAPIDTO(BaseModel):
    id: str = Field(alias="id")
    borrower_id: str = Field(alias="borrowerId")
    contact_id: str = Field(alias="contactId")
    priority: int = Field(alias="priority")
    is_active: bool = Field(alias="isActive")
    is_legal_representative: bool = Field(alias="isLegalRepresentative")
    relationship: str = Field(alias="relationship")
    ownership_pct: Optional[float] = Field(alias="ownershipPct")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateRelationshipDTO(BaseModel):
    borrower_id: str = Field(alias="borrowerId")
    contact_id: str = Field(alias="contactId")
    priority: Optional[int] = Field(alias="priority", default=None)
    relationship: str = Field(alias="relationship")
    is_legal_representative: Optional[bool] = Field(alias="isLegalRepresentative")
    is_active: Optional[bool] = Field(alias="isActive")
    ownership_pct: Optional[float] = Field(None, alias="ownershipPct")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateRelationshipDTO(BaseModel):
    priority: Optional[int] = Field(alias="priority", default=None)
    relationship: Optional[str] = Field(alias="relationship")
    is_legal_representative: Optional[bool] = Field(alias="isLegalRepresentative")
    is_active: Optional[bool] = Field(alias="isActive")
    ownership_pct: Optional[float] = Field(None, alias="ownershipPct")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class RelationshipSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "relationships", header_builder, renew_token, RelationshipAPIDTO.parse_obj(data))


class RelationshipAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "relationships", header_builder, renew_token, RelationshipAPIDTO.parse_obj(data))


class RelationshipsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=RelationshipSync,
                         retrieve_data_model=RelationshipAPIDTO,
                         create_data_model=CreateRelationshipDTO,
                         update_data_model=UpdateRelationshipDTO,
                         resource="relationships")


class RelationshipsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=RelationshipAsync,
                         retrieve_data_model=RelationshipAPIDTO,
                         create_data_model=CreateRelationshipDTO,
                         update_data_model=UpdateRelationshipDTO,
                         resource="relationships")



from typing import Optional,List
from pydantic import BaseModel, Field

class Rule(BaseModel):
    data_type: str = Field(alias="dataType")
    field: str = Field(alias="field")
    operator: str = Field(alias="operator")
    values: List[str] = Field(alias="values")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class RuleGroup(BaseModel):
    operator: str = Field(alias="operator")
    rules: list[Rule] = Field(alias="rules")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class RuleGroups(BaseModel):
    operator: str = Field(alias="operator")
    rule_group: RuleGroup = Field(alias="ruleGroup")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class RuleSet(BaseModel):
    name: str = Field(alias="name")
    rule_groups: list[RuleGroups] = Field(alias="ruleGroups")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class DPASegmentationAPIDTO(BaseModel):
    id: str = Field(alias="segmentationId")
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    rule_set: RuleSet = Field(alias="ruleSet")
    partner_id: str = Field(alias="partnerId")
    product_id: str = Field(alias="productId")
    created_at: str = Field(alias="createdAt")
    status: str = Field(alias="status")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)
    tenant: str = Field(alias="tenant")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class CreateDPASegmentationDTO(BaseModel):
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    rule_set: RuleSet = Field(alias="ruleSet")
    product_id: str = Field(alias="productId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True

class UpdateDPASegmentationDTO(BaseModel):
    name: str = Field(alias="name")
    description: Optional[str] = Field(alias="description", default=None)
    rule_set: RuleSet = Field(alias="ruleSet")
    product_id: str = Field(alias="productId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        populate_by_alias = True


class SegmentationBase:
    def __init__(self, base_url):
        self.base_url = base_url


class SegmentationAsync(SegmentationBase):
    data: DPASegmentationAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DPASegmentationAPIDTO):
        super().__init__(base_url)
        self.base_url = base_url
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.partner_id})"

class SegmentationSync(SegmentationBase):
    data: DPASegmentationAPIDTO

    def __init__(self, base_url, header_builder, renew_token, data: DPASegmentationAPIDTO):
        super().__init__(base_url)
        self._header_builder = header_builder
        self.renew_token = renew_token
        self.data = data

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data.partner_id})"


from typing import Optional


def build_headers(module, **kwargs):
    headers = {}
    if isinstance(module.altscore_client.user_token, str):
        user_token = module.altscore_client.user_token.replace("Bearer ", "")
        headers["Authorization"] = f"Bearer {user_token}"
    return headers



from pydantic import BaseModel, Field
from typing import Optional, List, Any, Dict, Union
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule
from altscore.borrower_central.model.accounting_amount import Amount

class DocumentDates(BaseModel):
    """
    Standardised date set for any accounting-related document.
    """
    issue_date: str = Field(
        alias="issueDate",
        description=(
            "Date the document was created / signed / CAE-stamped. "
            "Used for legal retention and tax‐book period."
        ),
    )
    posting_date: Optional[str] = Field(
        default=None,
        alias="postingDate",
        description=(
            "Date the document hit the general ledger (SAP “contable”). "
            "Controls accounting period close and revaluation runs."
        ),
    )
    due_date: Optional[str] = Field(
        default=None,
        alias="dueDate",
        description="Contractual deadline for payment; drives AP ageing and DSO.",
    )
    payment_date: Optional[str] = Field(
        default=None,
        alias="paymentDate",
        description=(
            "Actual settlement date when cash left or entered the company. "
            "Basis for cash-flow and realised FX gains/losses."
        ),
    )
    clearing_date: Optional[str] = Field(
        default=None,
        alias="clearingDate",
        description=(
            "Date the liability/receivable was cleared inside the ERP. "
            "May differ from payment_date (e.g. cheque deposited two days later)."
        ),
    )
    goods_receipt_date: Optional[str] = Field(
        default=None,
        alias="goodsReceiptDate",
        description="Physical arrival date of goods (GRN) for inventory documents.",
    )
    service_entry_date: Optional[str] = Field(
        default=None,
        alias="serviceEntryDate",
        description="Date a service was accepted (SES); used for accruals.",
    )
    tax_point_date: Optional[str] = Field(
        default=None,
        alias="taxPointDate",
        description=(
            "VAT ‘tax point’ / ‘date of supply’ when different from issue_date; "
            "relevant in UK/EU and some LATAM regimes."
        ),
    )
    value_date: Optional[str] = Field(
        default=None,
        alias="valueDate",
        description=(
            "Bank’s effective date for interest or FX on the payment "
            "(treasury forecasting and bank reconciliation)."
        ),
    )
    period_start: Optional[str] = Field(
        default=None,
        alias="periodStart",
        description="Start of the service/coverage period for periodic billing.",
    )
    period_end: Optional[str] = Field(
        default=None,
        alias="periodEnd",
        description="End of the service/coverage period for periodic billing.",
    )

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True

class AccountingDocumentAPIDTO(BaseModel):
    id: str = Field(alias="id")
    parent_document_id: Optional[str] = Field(alias="parentAccountingDocumentId", default=None)
    label: Optional[str] = Field(alias="label")
    key: str = Field(alias="key")
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    deal_id: Optional[str] = Field(alias="dealId", default=None)
    voucher: Optional[str] = Field(alias="voucher", default=None)
    reference_voucher: Optional[str] = Field(alias="referenceVoucher", default=None)
    producer_document_reference: Optional[str] = Field(alias="producerDocumentReference", default=None)
    receiver_document_reference: Optional[str] = Field(alias="receiverDocumentReference", default=None)
    is_outbound: bool = Field(alias="isOutbound")
    document_dates: Optional[DocumentDates] = Field(alias="documentDates", default=None)
    tags: List[str] = Field(alias="tags", default=[])
    raw_content: Optional[str] = Field(alias="rawContent", default=None)
    content: Optional[Dict[str, Any]] = Field(alias="content", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)
    amount: Optional[Amount] = Field(alias="amount", default=None)
    confidence_score: Optional[float] = Field(alias="confidenceScore", default=None)
    has_attachments: bool = Field(alias="hasAttachments")
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateAccountingDocumentDTO(BaseModel):
    key: str = Field(alias="key")
    is_outbound: bool = Field(alias="isOutbound")
    parent_document_id: Optional[str] = Field(alias="parentAccountingDocumentId", default=None)
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    deal_id: Optional[str] = Field(alias="dealId", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    voucher: Optional[str] = Field(alias="voucher", default=None)
    reference_voucher: Optional[str] = Field(alias="referenceVoucher", default=None)
    producer_document_reference: Optional[str] = Field(alias="producerDocumentReference", default=None)
    receiver_document_reference: Optional[str] = Field(alias="receiverDocumentReference", default=None)
    document_dates: Union[Dict, DocumentDates, None] = Field(alias="documentDates", default=None)
    tags: Optional[List[str]] = Field(alias="tags", default=None)
    raw_content: Optional[str] = Field(alias="rawContent", default=None)
    content: Optional[Dict[str, Any]] = Field(alias="content", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)
    amount: Optional[Amount] = Field(alias="amount", default=None)
    confidence_score: Optional[float] = Field(alias="confidenceScore", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateAccountingDocumentDTO(BaseModel):
    is_outbound: Optional[bool] = Field(alias="isOutbound", default=None)
    parent_document_id: Optional[str] = Field(alias="parentAccountingDocumentId", default=None)
    borrower_id: Optional[str] = Field(alias="borrowerId", default=None)
    deal_id: Optional[str] = Field(alias="dealId", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    voucher: Optional[str] = Field(alias="voucher", default=None)
    reference_voucher: Optional[str] = Field(alias="referenceVoucher", default=None)
    producer_document_reference: Optional[str] = Field(alias="producerDocumentReference", default=None)
    receiver_document_reference: Optional[str] = Field(alias="receiverDocumentReference", default=None)
    document_dates: Union[Dict, DocumentDates, None] = Field(alias="documentDates", default=None)
    tags: Optional[List[str]] = Field(alias="tags", default=None)
    raw_content: Optional[str] = Field(alias="rawContent", default=None)
    content: Optional[Dict[str, Any]] = Field(alias="content", default=None)
    metadata: Optional[Dict[str, Any]] = Field(alias="metadata", default=None)
    amount: Optional[Amount] = Field(alias="amount", default=None)
    confidence_score: Optional[float] = Field(alias="confidenceScore", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class AccountingDocumentSync(GenericSyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "accounting/documents", header_builder,
                         renew_token, AccountingDocumentAPIDTO.parse_obj(data))


class AccountingDocumentAsync(GenericAsyncResource):

    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "accounting/documents", header_builder,
                         renew_token, AccountingDocumentAPIDTO.parse_obj(data))


class AccountingDocumentsSyncModule(GenericSyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=AccountingDocumentSync,
                         retrieve_data_model=AccountingDocumentAPIDTO,
                         create_data_model=CreateAccountingDocumentDTO,
                         update_data_model=UpdateAccountingDocumentDTO,
                         resource="accounting/documents")


class AccountingDocumentsAsyncModule(GenericAsyncModule):

    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=AccountingDocumentAsync,
                         retrieve_data_model=AccountingDocumentAPIDTO,
                         create_data_model=CreateAccountingDocumentDTO,
                         update_data_model=UpdateAccountingDocumentDTO,
                         resource="accounting/documents")



import httpx
from altscore.common.http_errors import raise_for_status_improved, retry_on_401, retry_on_401_async
from pydantic import BaseModel, Field
from typing import Optional, List, Dict
from altscore.borrower_central.model.generics import GenericSyncResource, GenericAsyncResource, \
    GenericSyncModule, GenericAsyncModule


# DTO for assets
class AssetDTO(BaseModel):
    """Data transfer object for assets"""
    id: str = Field(alias="id")
    deal_id: str = Field(alias="dealId")
    external_id: Optional[str] = Field(alias="externalId", default=None)
    label: Optional[str] = Field(alias="label", default=None)
    description: Optional[str] = Field(alias="description", default=None)
    group: Optional[str] = Field(alias="group", default=None)
    created_at: str = Field(alias="createdAt")
    updated_at: Optional[str] = Field(alias="updatedAt", default=None)
    has_attachments: bool = Field(alias="hasAttachments")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class CreateAssetRequest(BaseModel):
    """Model for creating a new asset"""
    deal_id: str = Field(alias="dealId")
    label: str = Field(alias="label")
    description: Optional[str] = Field(alias="description", default=None)
    external_id: Optional[str] = Field(alias="externalId", default=None)
    group: Optional[str] = Field(alias="group", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class UpdateAssetRequest(BaseModel):
    """Model for updating an asset"""
    label: Optional[str] = Field(alias="label", default=None)
    description: Optional[str] = Field(alias="description", default=None)

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


class ExternalIdRequest(BaseModel):
    """Model for setting an external ID"""
    external_id: str = Field(alias="externalId")

    class Config:
        populate_by_name = True
        allow_population_by_field_name = True
        allow_population_by_alias = True


# Resource classes for assets
class AssetSync(GenericSyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "assets", header_builder, renew_token, AssetDTO.parse_obj(data))


class AssetAsync(GenericAsyncResource):
    def __init__(self, base_url, header_builder, renew_token, data: Dict):
        super().__init__(base_url, "assets", header_builder, renew_token, AssetDTO.parse_obj(data))


# Module for assets - synchronous
class AssetsSyncModule(GenericSyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         sync_resource=AssetSync,
                         retrieve_data_model=AssetDTO,
                         create_data_model=CreateAssetRequest,
                         update_data_model=UpdateAssetRequest,
                         resource="assets")

    @retry_on_401
    def set_external_id(self, asset_id: str, external_id: str):
        """
        Set an external ID for an asset
        
        Args:
            asset_id: The ID of the asset
            external_id: The external ID to set
            
        Returns:
            None
        """
        request_data = ExternalIdRequest(externalId=external_id)

        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.put(
                f"/v1/assets/{asset_id}/external-id",
                json=request_data.dict(by_alias=True, exclude_none=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401
    def query_by_deal_id(self, deal_id: str, page: int = 1, per_page: int = 10):
        """
        Find assets by deal ID
        
        Args:
            deal_id: The ID of the deal to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            Dict with assets and pagination info
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/assets",
                params={
                    "deal-id": deal_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()

    @retry_on_401
    def retrieve_by_external_id(self, external_id: str) -> Optional[AssetSync]:
        """
        Retrieve an asset by its external ID
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/assets",
                params={"external-id": external_id},
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            assets = response.json()
            if assets and len(assets) > 0:
                return AssetSync(
                    base_url=self.altscore_client._borrower_central_base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    data=assets[0]
                )
            return None

    @retry_on_401
    def query_by_group(self, group: str, page: int = 1, per_page: int = 10):
        """
        Find assets by group

        Args:
            group: The group key to filter by
            page: Page number for pagination
            per_page: Number of results per page

        Returns:
            Dict with assets and pagination info
        """
        with httpx.Client(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = client.get(
                "/v1/assets",
                params={
                    "group": group,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()


# Module for assets - asynchronous
class AssetsAsyncModule(GenericAsyncModule):
    def __init__(self, altscore_client):
        super().__init__(altscore_client,
                         async_resource=AssetAsync,
                         retrieve_data_model=AssetDTO,
                         create_data_model=CreateAssetRequest,
                         update_data_model=UpdateAssetRequest,
                         resource="assets")

    @retry_on_401_async
    async def set_external_id(self, asset_id: str, external_id: str):
        """
        Set an external ID for an asset
        
        Args:
            asset_id: The ID of the asset
            external_id: The external ID to set
            
        Returns:
            None
        """
        request_data = ExternalIdRequest(externalId=external_id)

        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.put(
                f"/v1/assets/{asset_id}/external-id",
                json=request_data.dict(by_alias=True, exclude_none=True),
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return None

    @retry_on_401_async
    async def query_by_deal_id(self, deal_id: str, page: int = 1, per_page: int = 10):
        """
        Find assets by deal ID
        
        Args:
            deal_id: The ID of the deal to filter by
            page: Page number for pagination
            per_page: Number of results per page
            
        Returns:
            Dict with assets and pagination info
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/assets",
                params={
                    "deal-id": deal_id,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()

    @retry_on_401_async
    async def retrieve_by_external_id(self, external_id: str) -> Optional[AssetAsync]:
        """
        Retrieve an asset by its external ID
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/assets",
                params={"external-id": external_id},
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            assets = response.json()
            if assets and len(assets) > 0:
                return AssetAsync(
                    base_url=self.altscore_client._borrower_central_base_url,
                    header_builder=self.build_headers,
                    renew_token=self.renew_token,
                    data=assets[0]
                )
            return None

    @retry_on_401_async
    async def query_by_group(self, group: str, page: int = 1, per_page: int = 10):
        """
        Find assets by group

        Args:
            group: The group key to filter by
            page: Page number for pagination
            per_page: Number of results per page

        Returns:
            Dict with assets and pagination info
        """
        async with httpx.AsyncClient(base_url=self.altscore_client._borrower_central_base_url) as client:
            response = await client.get(
                "/v1/assets",
                params={
                    "group": group,
                    "page": page,
                    "per-page": per_page
                },
                headers=self.build_headers(),
                timeout=120,
            )
            raise_for_status_improved(response)
            return response.json()
